/**
 * @file object.cpp
 * Implementation of Object class.
 *
 * Copyright 2003 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "object.h"
#include "msmodel.h"
#include "shapemodel.h"
#include "utils.h"

// add a loader function for this object
static RegisterLoader dummy( Object::get_type_str_static(), &Object::create );

int Object::load( dWorldID world_id, dSpaceID space_id, const CL_DomElement& element, bool load_parent )
{
    if ( element.has_attribute( "position" ) ) {
        set_position( get_vec3( element, "position" ) );
    }
    else {
        set_position( Vec3r( 0.0 ) );
    }

    if ( element.has_attribute( "forward" ) && element.has_attribute( "up" ) ) {
        // objects point down the +z axis by default.
        Vec3r z_axis = get_vec3( element, "forward" );
        Vec3r y_axis = get_vec3( element, "up" );
        Vec3r x_axis = y_axis ^ z_axis;

        dMatrix3 rot_mat;
        dRFrom2Axes( rot_mat,
                x_axis[ 0 ], x_axis[ 1 ], x_axis[ 2 ],
                y_axis[ 0 ], y_axis[ 1 ], y_axis[ 2 ] );
        set_orientation_mat( rot_mat );
    }
    else {
        dQuaternion dq;
        dQSetIdentity( dq );
        set_orientation_quat( dq );
    }

    if ( element.has_attribute( "model" ) ) {
        string filename = get_string( element, "model" );
        _model = new MSModel( filename );
    }
    else if ( element.has_attribute( "shape" ) ) {
        Shape shape = get_shape( element, "shape" );
        _model = new ShapeModel( shape, Vec3r( 1.0 ) );
    }
    else {
        _model = new Model;
    }

    if ( element.has_attribute( "texture" ) ) {
        string filename = get_string( element, "texture" );
        Material mat = grey.copy_with_texture( filename );
        _model->set_material( mat );
    }

    if ( element.has_attribute( "scale" ) ) {
        set_scale( get_vec3( element, "scale" ) );
    }
    else {
        set_scale( Vec3r( 1.0 ) );
    }

    if ( element.has_attribute( "mass" ) ) {
        set_mass( get_double( element, "mass" ) );
    }
    else {
        set_mass( 1.0 );
    }

    if ( element.has_attribute( "on" ) ) {
        set_on( get_bool( element, "on" ) );
    }
    else {
        set_on( false );
    }

    if ( element.has_attribute( "stationary" ) ) {
        set_stationary( get_bool( element, "stationary" ) );
    }
    else {
        set_stationary( false );
    }

    if ( element.has_attribute( "movable" ) ) {
        set_movable( get_bool( element, "movable" ) );
    }
    else {
        set_movable( true );
    }

    if ( element.has_attribute( "rotatable" ) ) {
        Vec3r rotatable_vec = get_vec3( element, "rotatable" );
        bool rotatable[ 3 ];
        rotatable[ 0 ] = bool( rotatable_vec[ 0 ] );
        rotatable[ 1 ] = bool( rotatable_vec[ 1 ] );
        rotatable[ 2 ] = bool( rotatable_vec[ 2 ] );

        set_rotatable( rotatable );
    }
    else {
        set_rotatable( true, true, true );
    }

    _init_on = _on;

    return 0;
}

Object::Object( dWorldID world_id, dSpaceID space_id, const CL_DomElement& element )
{
    _geom_id = 0;
    _body_id = dBodyCreate( world_id );
    dBodySetData( _body_id, this );

    load( world_id, space_id, element, false );
}

Object::~Object( void )
{
    // remove this object from ode
    if ( _body_id ) {
        dBodyDestroy( _body_id );
    }
    if ( _geom_id ) {
        dGeomDestroy( _geom_id );
    }

    // I assume the object will gain control of the model, and can delete
    // it at will.
	if( _model ) {
		delete _model;
	}
}

EffectHandler Object::get_effect_handler( Effect *effect )
{
//    if ( IS_PHYSICAL_EFFECT( effect ) && !is_ethereal() ) {
//        return (EffectHandler)&Object::_handle_physical_effect;
//    }
    return NULL; // no handler for this effect
}

void Object::render( bool selected )
{
    glPushMatrix();

    apply_transform();

    _model->render();

    glPopMatrix();
}

void Object::draw_decor( Color color )
{
    glPushMatrix();

    apply_transform();

    _model->draw_decor( color );

    glPopMatrix();
}

void Object::apply_transform( void )
{
    Vec3r pos = dBodyGetPosition( _body_id );

    glTranslated( pos[ 0 ], pos[ 1 ], pos[ 2 ] );

#if 1
    dReal rot[ 16 ];

    dMatrix3ToOpenGL( rot, dBodyGetRotation( _body_id ) );

    glMultMatrixf( rot );
#else
    // I'm pretty sure quaternions are faster
    // update: unfortunately, they break sometimes.
    const dReal *quat = dBodyGetQuaternion( _body_id );
    double rot_angle = RAD_TO_DEG( 2.0 * acos( quat[0] ) );

    glRotatef( rot_angle, quat[ 1 ], quat[ 2 ], quat[ 3 ] );
#endif

    glScaled( _scale[ 0 ], _scale[ 1 ], _scale[ 2 ] );
}

void Object::start_sim( void )
{
    dRealPtrToVec3( _init_position, dBodyGetPosition( _body_id ) );

    dRealPtrTodQuaternion( _init_orientation, dBodyGetQuaternion( _body_id ) );

    _init_on = _on;
}

void Object::stop_sim( void )
{
    dBodySetPosition( _body_id, _init_position[ 0 ], _init_position[ 1 ], _init_position[ 2 ] );

    dBodySetQuaternion( _body_id, _init_orientation );

    dBodySetLinearVel( _body_id, 0.0, 0.0, 0.0 );
    dBodySetAngularVel( _body_id, 0.0, 0.0, 0.0 );

    _on = _init_on;

    // are these necessary?
    dBodySetForce( _body_id, 0.0, 0.0, 0.0 );
    dBodySetTorque( _body_id, 0.0, 0.0, 0.0 );
}

void Object::_setup_geom( void )
{
    dBodySetMass( _body_id, &_dmass );

    dGeomSetBody( _geom_id, _body_id );
}

void Object::collide( dContact* contact )
{

}

string Object::get_info( ) const
{
    ostringstream out;

    out << "type=\"" << get_type_str() << "\" ";
    out << get_model()->get_info() << " ";
    out << endl;
    out << "position=\"" << get_position() << "\" ";
    out << "forward=\"" << get_forward_vec() << "\" ";
    out << "up=\"" << get_up_vec() << "\" ";
    out << endl;
    out << "scale=\"" << get_scale() << "\" ";
    out << "mass=\"" << get_mass() << "\" ";
    out << endl;
    out << "on=\"" << to_string( is_on() ) << "\" ";
    out << "stationary=\"" << to_string( is_stationary() ) << "\" ";
    out << endl;
    out << "movable=\"" << to_string( is_movable() ) << "\" ";
    out << "rotatable=\""
        << ( is_rotatable( 0 ) ? 1 : 0 ) << " "
        << ( is_rotatable( 1 ) ? 1 : 0 ) << " "
        << ( is_rotatable( 2 ) ? 1 : 0 ) << "\" ";

    return out.str();
}


typedef map< string, Object::LoadFunc* > LoaderMap;
static LoaderMap* loader_map = NULL;

RegisterLoader::RegisterLoader( string str, Object::LoadFunc* func )
{
    if ( !loader_map ) {
        loader_map = new LoaderMap;
    }

    (*loader_map)[ str ] = func;
}

Object::LoadFunc* RegisterLoader::get_loader( string str )
{
    assert( loader_map );
    return (*loader_map)[ str ];
}
