/**
 * @file object.h
 * Header for Object class.
 *
 * Copyright 2003 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _OBJECT_H
#define _OBJECT_H

#include "main.h"
#include "actor.h"
#include "odeutils.h"
#include "gfxutils.h"

#include <assert.h>
#include <sstream>
#include <string>
using namespace std;

#include <ClanLib/core.h>

#include <gfx/mat4.h>
#include <gfx/quat.h>
#include <gfx/gl.h>

#include <ode/ode.h>

class Object;
class Model;
class Effect;

/**
 * A method pointer for effect handlers.
 *
 * What gets called when the Object recieves a certain effect.
 */
typedef void (Object::*EffectHandler)(Effect *);

/**
 * A physical, visible object.
 *
 * Objects are physical things in the scene, and interact with each other
 * via effects and collisions.  To receive effects, an Object must implement
 * get_effect_handler(), and return the function it wants called whenever
 * it receives a effect of a given type.  Collisions are handled by ode.
 */
class Object : public Actor {
public:

    typedef Object* LoadFunc( dWorldID world_id, dSpaceID space_id, const CL_DomElement& );

    static string get_type_str_static( ) { return "object"; }

    static Object* create( dWorldID world_id, dSpaceID space_id, const CL_DomElement& element )
        { return new Object( world_id, space_id, element ); }

    /**
     * Set this object's parameters from the given DOM element (xml).
     *
     * @return 0 on success, -1 on failure.
     */
    virtual int load( dWorldID world_id, dSpaceID space_id, const CL_DomElement& element, bool load_parent = true );

    /**
     * Object constructor.
     *
     * @param world_id ode world in which to place this object's body
     * @param space_id ode space in which to place this objec's geometry
     * @param element  xml element which holds the parameters to be used in
     *                 creating the object
     */
    Object( dWorldID world_id, dSpaceID space_id, const CL_DomElement& element );

    /**
     * Object destructor
     *
     * Assumes the Object has gained control of its model and is able to call
     * delete on it.
     */
    virtual ~Object( void );

    /**
     * Return the effect handler for a particular type of effect.
     *
     * Determine what kind of Effect effect is, and returns a
     * method pointer that will handle applications of that effect.
     *
     * @param effect the effect to find the handler for
     *
     * @return the handler (method pointer) for the given effect
     */
    virtual EffectHandler get_effect_handler( Effect *effect );

    /**
     * Render the object to the screen.
     *
     * Currently just calls the model's render() function.
     *
     * @param selected unused
     */
    virtual void render( bool selected = false );

    /**
     * Draw a decoration around the object to show its status.
     *
     * @param color the RGB color of the decoration
     */
    void draw_decor( Color color );

    /**
     * Apply glTranslate, glRotate, and glScale transformations to the model.
     */
    virtual void apply_transform( void );

    /**
     * Advance the object's state.
     *
     * @param t amount of time (in seconds) to advance the object's state
     */
    virtual void tick( double t ) { }

    /**
     * Notify the object that the simulation has started.
     *
     * The object will need to save its current state so that it can be reset
     * when the simulation stops.
     */
    virtual void start_sim( void );

    /**
     * Notify the object that the simulation has stopped.
     *
     * The object will need to reset its state to what it was before the
     * simulation started.
     */
    virtual void stop_sim( void );

    /**
     * Notify the object that it has been collided with.
     *
     * Allows the object to set salient properties of the contact joint.
     *
     * @param contact ode dContact* which determines the properties for the
                      contact joint generated by the current collision
     */
    virtual void collide( dContact* contact );

    //@{
    /** @name Accessors
     *
     * Accessor functions.
     */
    void set_id( int id ) { _id = id; }
    void set_position( const Vec3r& pos ) { dBodySetPosition( _body_id, pos[ 0 ], pos[ 1 ], pos[ 2 ] ); }
    void set_orientation_mat( const dMatrix3 orientation ) { dBodySetRotation( _body_id, orientation ); }
    void set_orientation_quat( const dQuaternion orientation ) { dBodySetQuaternion( _body_id, orientation ); }
    void set_mass( float m ) { _mass = m; }
    void set_scale( const Vec3r s ) { _scale = s; }
    void set_stationary( const bool s ) { if( ( _stationary = s ) ) dBodyDisable( _body_id ); else dBodyEnable( _body_id ); }
    void set_movable( const bool s ) { _movable = s; }
    void set_rotatable( bool x, bool y, bool z )
        { _rotatable[0] = x; _rotatable[1] = y; _rotatable[2] = z; }
    void set_rotatable( const bool s[3] )
        { _rotatable[0] = s[0]; _rotatable[1] = s[1]; _rotatable[2] = s[2]; }
    void set_ethereal( const bool e ) { _ethereal = e; }
    void set_on( const bool o ) { _on = o; }

    int get_id( ) const { return _id; }
    Vec3r get_position( void ) const { const float* pos = dBodyGetPosition( _body_id ); return Vec3r( pos[ 0 ], pos[ 1 ], pos[ 2 ] ); }
    void get_orientation_mat( dMatrix3 dmat3 ) const { dRealPtrTodMatrix3( dmat3, dBodyGetRotation( _body_id ) ); }
    Mat4 get_orientation_mat( ) const { Mat4 m; dMatrix3ToMat4( m, dBodyGetRotation( _body_id ) ); return m; }
    void get_orientation_quat( dQuaternion dquat ) const { dRealPtrTodQuaternion( dquat, dBodyGetQuaternion( _body_id ) ); }
    Quat get_orientation_quat( ) const { Quat q; dQuaternion_to_quat( q, dBodyGetQuaternion( _body_id ) ); return q; }
    Vec3r get_forward_vec( ) const { return get_orientation_mat() * Vec3r( 0, 0, 1 ); }
    Vec3r get_up_vec( ) const { return get_orientation_mat() * Vec3r( 0, 1, 0 ); }
    virtual float get_mass( void ) const { dMass dm; dBodyGetMass( _body_id, &dm ); return (float) dm.mass; }
    virtual Vec3r get_scale( void ) const { return _scale; }
    virtual bool is_stationary( void ) const { return _stationary; }
    virtual bool is_movable( void ) const { return _movable; }
    virtual bool is_rotatable( int i ) const { return _rotatable[i]; }
    virtual bool is_ethereal( void ) const { return _ethereal; }
    virtual bool is_on( void ) const { return _on; }
    virtual Model *get_model( ) const { return _model; }
    virtual string get_type_str( ) const { return Object::get_type_str_static(); }
    virtual string get_info( ) const;

    Vec3r get_init_position( void ) const { return _init_position; }
    void get_init_orientation_mat( dMatrix3 dmat3 ) const { dQtoR( _init_orientation, dmat3 ); }
    void get_init_orientation_quat( dQuaternion dquat ) const { dRealPtrTodQuaternion( dquat, _init_orientation ); }
    bool is_init_on( void ) const { return _init_on; }

    const dBodyID get_body_id( void ) { return _body_id; }
    const dGeomID get_geom_id( void ) { return _geom_id; }

    void enable_body( void ) { dBodyEnable( _body_id ); }
    void disable_body( void ) { dBodyDisable( _body_id ); }

    //@}

protected:

    /**
     * Populate the object's ode geometry related structures.
     */
    void _setup_geom( void );

    dBodyID _body_id;
    dGeomID _geom_id;

    //@{
    /** @name Object's simulation state
    */
    Vec3r _init_position;
    dQuaternion _init_orientation;
    bool _init_on;
    //@}

    dMass _dmass;               ///< mass as an ode struct
    float _mass;                ///< mass as a constant

    Model* _model;              ///< my graphical representation

    Vec3r _scale;                ///< the scaling factor

    int _id;                    ///< unique identifier

    bool _stationary;           ///< does the physical simulation affect me?
    bool _movable;              ///< can the user move me?
    bool _rotatable[ 3 ];       ///< can the user rotate me about x,y,z?
    bool _ethereal;             ///< can objects collide with me / can I overlap objects
    bool _on;                   ///< am i activated?
};

class RegisterLoader {
public:

    RegisterLoader( string str, Object::LoadFunc* func );

    static Object::LoadFunc* get_loader( string str );
};

#endif // header guard
