/**
 * @file actors/worldbox.cpp
 * Implementation of Actors/worldbox class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "worldbox.h"
#include "model.h"
#include "material.h"

#include <sstream>

WorldBox::WorldBox( dSpaceID space_id, Vec3 size, const Material& wall_mat, const Material& background_mat ) :
    _size( size ),
    _min( -_size / 2.0 ),
    _max( size / 2.0 ),
    _background_ds( 0.0 ),
    _background_dt( 0.0 )

{
    _wall = new Material( wall_mat );
    _background = new Material( background_mat );

    _wall->reload_texture();
    _background->reload_texture();

    // attach the bounding planes to the static environment (i.e. they dont
    // need rigid body id's)
    _geom_id[ 0 ] = dCreatePlane( space_id,  1,  0,  0,  _min[ 0 ] );
    _geom_id[ 1 ] = dCreatePlane( space_id, -1,  0,  0, -_max[ 0 ] );
    _geom_id[ 2 ] = dCreatePlane( space_id,  0,  1,  0,  _min[ 1 ] );
    _geom_id[ 3 ] = dCreatePlane( space_id,  0, -1,  0, -_max[ 1 ] );
    _geom_id[ 4 ] = dCreatePlane( space_id,  0,  0,  1,  _min[ 2 ] );
    _geom_id[ 5 ] = dCreatePlane( space_id,  0,  0, -1, -_max[ 2 ] );
}

WorldBox::~WorldBox( void )
{
    for ( int i = 0; i < 6; i++ ) {
        dGeomDestroy( _geom_id[ i ] );
    }

	if( _wall ) {
		delete _wall;
	}

	if( _background ) {
		delete _background;
	}
}

void WorldBox::render( Vec3r camera_position )
{
    glMatrixMode( GL_MODELVIEW );
    glPushMatrix();

    _wall->make_current();

    // figure out which sides of the box the camera is on, and don't draw
    // those sides
    bool draw_faces[ 6 ] = { true, true, true, true, true, true };

    if ( camera_position[ 0 ] < _min[ 0 ] ) {
        draw_faces[ CUBE_LEFT ] = false;
    }
    else if ( camera_position[ 0 ] > _max[ 0 ] ) {
        draw_faces[ CUBE_RIGHT ] = false;
    }
    if ( camera_position[ 1 ] < _min[ 1 ] ) {
        draw_faces[ CUBE_BOTTOM ] = false;
    }
    else if ( camera_position[ 1 ] > _max[ 1 ] ) {
        draw_faces[ CUBE_TOP ] = false;
    }
    if ( camera_position[ 2 ] < _min[ 2 ] ) {
        draw_faces[ CUBE_FRONT ] = false;
    }
    else if ( camera_position[ 2 ] > _max[ 2 ] ) {
        draw_faces[ CUBE_BACK ] = false;
    }

    Model::draw_solid_cube( _min, _max, draw_faces );

    glMatrixMode( GL_MODELVIEW );
    glPopMatrix();


    // clouds

    glMatrixMode( GL_TEXTURE );
    glPushMatrix();

    _background->make_current();
    _background_ds += 0.0003;
    _background_dt += 0.0001;
    glTranslated( _background_ds, _background_dt, 0.0 );
    glScaled( 10.0, 10.0, 10.0 );

    glMatrixMode( GL_MODELVIEW );
    glPushMatrix();
    glLoadIdentity();

    glRotated( 45.0, 0.0, 0.0, 1.0 );
    glScaled( 1.0, 2.0, 1.0 );
    glRotated( -45.0, 0.0, 0.0, 1.0 );

    glBegin( GL_QUADS );
    const double size = 200.0;
    const double dist = 200.0;

    glNormal3d( 0, 0, -1 );
    glTexCoord2f( 0, 1 );
    glVertex3d( -size, -size, -dist );
    glTexCoord2f( 0, 0 );
    glVertex3d( -size, size, -dist );
    glTexCoord2f( 1, 0 );
    glVertex3d( size, size, -dist );
    glTexCoord2f( 1, 1 );
    glVertex3d( size, -size, -dist );

    glEnd();

    glMatrixMode( GL_TEXTURE );
    glPopMatrix();

    glMatrixMode( GL_MODELVIEW );
    glPopMatrix();
}

string WorldBox::get_info( void ) const
{
    ostringstream out;

    out << "size=\"" << _size << "\" ";
    out << "wall=\"" << _wall->texture_filename << "\" ";
    out << "background=\"" << _background->texture_filename << "\" ";

    return out.str();
}
