/**
 * @file gamestate.cpp
 * Implementation of Gamestate class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "gamestate.h"
#include "level.h"
#include "camera.h"
#include "object.h"

GameState::GameState( CL_GraphicContext* gc ) :
    _current_level( NULL ),
    _current_state( MAIN ),
    _previous_state( MAIN )
{
    if( !gc ) {
        cerr << "GameState::GameState: NULL GraphicContext!" << endl;
    }

    _glstate = new CL_OpenGLState( gc );

    _gl_init();
}

GameState::~GameState( void )
{
    if( _current_level ) {
        delete _current_level;
    }

	if( _glstate ) {
		delete _glstate;
	}
}

void GameState::_gl_init( void )
{
    _glstate->set_active();

    glClearColor( 0.078f, 0.361f, 0.753f, 1.0f );
    glClearDepth( 1.0 );

    glDepthFunc( GL_LESS );
    glEnable( GL_DEPTH_TEST );

    glShadeModel( GL_SMOOTH );
    glEnable( GL_NORMALIZE );

    glEnable( GL_LINE_SMOOTH );

    glEnable( GL_LIGHTING );
    glLightModelfv( GL_LIGHT_MODEL_AMBIENT, Vec4f( 0.3f, 0.3f, 0.3f, 1.0f ) );

    const Vec4f light0_pos( 0.0f, 10.0f, 0.0f, 0.0f );
    const Vec4f light0_rgb( 0.5f, 0.5f, 0.5f, 1.0f );

    glLightfv( GL_LIGHT0, GL_POSITION, light0_pos );
    glLightfv( GL_LIGHT0, GL_DIFFUSE, light0_rgb );
    glLightfv( GL_LIGHT0, GL_SPECULAR, light0_rgb );
    glEnable( GL_LIGHT0 );

    const Vec4f light1_pos( 0.0f, 5.0f, 10.0f, 0.0f );
    const Vec4f light1_rgb( 0.5f, 0.5f, 0.5f, 1.0f );

    glLightfv( GL_LIGHT1, GL_POSITION, light1_pos );
    glLightfv( GL_LIGHT1, GL_DIFFUSE, light1_rgb );
    glLightfv( GL_LIGHT1, GL_SPECULAR, light1_rgb );
    glEnable( GL_LIGHT1 );

    glEnable( GL_TEXTURE_2D );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
}

void GameState::change_state( const State state )
{
    _previous_state = _current_state;

    _current_state = state;

    // transition logic
    if ( _current_level ) {
        if ( _current_state == RUN_SIM && _previous_state == GAMEPLAY ) {
            _current_level->start_sim();
        }
        else if ( _current_state == GAMEPLAY && _previous_state == RUN_SIM ) {
            _current_level->stop_sim();
        }
    }
}

void GameState::load_level( const string name )
{
    // check if we're trying to reload the already loaded level instead of
    // blindly reloading the level?
    if ( _current_level ) {
        delete _current_level;
    }

    _current_level = new Level( name );
    _current_level->load();

    int matrix_mode;

    glGetIntegerv( GL_MATRIX_MODE, &matrix_mode );

    _glstate->set_active();

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();

    _current_level->get_camera()->apply_projection();

    glMatrixMode( matrix_mode );
}

void GameState::tick( double t )
{
    if ( _current_level && _current_state == RUN_SIM ) {
        _current_level->tick( t );

        // now that effects are into ode, check for collisions and reposition the objects
        dSpaceCollide( _current_level->get_space_id(), this, &GameState::_static_near_callback );

        dWorldStep( _current_level->get_world_id(), dReal( t ) );

        dJointGroupEmpty( _current_level->get_contact_group() );

        // check if the user beat the level in this tick
        if ( _current_level->level_completed() ) {
            if( true ) {
                // there is another level
                change_state( END_LEVEL );
            }
            else {
                // that was the last level
                change_state( END_GAME );
            }
        }

    }

    // any other per-frame checking the gamestate needs to do...
}

void GameState::render( void )
{
    _glstate->set_active();
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    if ( _current_level && ( _current_state == RUN_SIM ||
                             _current_state == GAMEPLAY ||
                             _current_state == LEVEL_LOADING ) ) {

        _current_level->render( _current_state == GAMEPLAY );
    }
    else {
        // render a background scene
    }
}

void GameState::reset_level( void )
{
    if ( !_current_level ) {
        cerr << "GameState::reset_level: attempt to reset a NULL level!" << endl;
        return;
    }

    string pathname = _current_level->get_pathname();

    load_level( pathname );

    _current_level->reset_level();

    change_state( GAMEPLAY );
}

void GameState::_static_near_callback( void* data, dGeomID o1, dGeomID o2 )
{
    ((GameState*) data)->_near_callback( o1, o2 );
}

void GameState::_near_callback( dGeomID o1, dGeomID o2 )
{
    const unsigned long N = 6;
    dContact contacts[ N ];

    const unsigned long n = dCollide( o1, o2, N, &contacts[ 0 ].geom, sizeof( dContact ) );

    dWorldID world_id = _current_level->get_world_id();
    dJointGroupID contact_group = _current_level->get_contact_group();

    for ( unsigned long i = 0; i < n; ++i ) {
        dBodyID body_id1 = dGeomGetBody( contacts[ i ].geom.g1 );
        dBodyID body_id2 = dGeomGetBody( contacts[ i ].geom.g2 );
        Object* object1 = (Object*) ( body_id1 ? dBodyGetData( body_id1 ) : NULL );
        Object* object2 = (Object*) ( body_id2 ? dBodyGetData( body_id2 ) : NULL );

        contacts[ i ].surface.mode = dContactBounce;
        contacts[ i ].surface.mu = 250;
        contacts[ i ].surface.bounce = 0.5f;
        contacts[ i ].surface.bounce_vel = 0.05f;

        // allow the objects to modify the contact joint or set themselves up
        // as static parts of the environment
        if ( object1 ) {
            object1->collide( &contacts[ i ] );

            if( object1->is_stationary() ) {
                body_id1 = 0;
            }
        }

        if ( object2 ) {
            object2->collide( &contacts[ i ] );

            if( object2->is_stationary() ) {
                body_id2 = 0;
            }
        }

        // only attach the joint if at least one of the object's are not
        // stationary
        if( ( body_id1 != 0 || body_id2 != 0 ) ) {
            dJointID c = dJointCreateContact( world_id, contact_group, &contacts[ i ] );
            dJointAttach( c, body_id1, body_id2 );
        }
    }
}
