/**
 * @file gamestate.h
 * Header file for Gamestate class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _GAMESTATE_H
#define _GAMESTATE_H

#include "main.h"

#include <string>
using namespace std;

#include <ClanLib/gl.h>

#include <gfx/vec4.h>

#include <ode/ode.h>

class Level;

/**
 * Holds the currently loaded level, game state information, and opengl state.
 *
 * Other classes depend on GameState to determine what action(s) they should 
 * take in the current game loop.  Action(s) taken on game state transitions 
 * should be done in this class though.  It is also responsible for making sure 
 * that the correct openglstate is active before the world is rendered. 
 *
 * @see GUI
 * @see InputHandler
 */
class GameState {
public:
    
    /**
     * Different states the game can be in. 
     *
     * The order of this enum affects the order in which menus are specified
     * in the GUI.
     *
     * @see GUI::GUI
     */
    enum State {
        MAIN = 0,               ///< main menu
        LEVEL_LOADING,          ///< loading level, display level description
        GAMEPLAY,               ///< user is able to interact with the level
        END_LEVEL,              ///< the level has been completed
        END_GAME,               ///< the game has been completed
        LEVEL_SELECT,           ///< user selects a level to play
        HELP,                   ///< help menu describing controls
        OPTIONS_VIDEO_SOUND,    ///< change video and sound options
        OPTIONS_CONTROLS,       ///< change control options
        ESC_DIALOG,             ///< esc dialog displable during gameplay
        RUN_SIM                 ///< the simulation is running
    };

    /**
     * GameState constructor.
     *
     * @param gc the opengl context created by the GUI's window
     */
    GameState( CL_GraphicContext* gc );

    ~GameState( void );

    /**
     * Update the game's state and take care of any transition logic.
     *
     * @param state the new state of the game
     */
    void change_state( const State state );

    /**
     * Load a new level, deleting an old level if it exists.
     *
     * Does initial setup of opengl state (modelview and projection matrices).
     *
     * @param filename relative path + filename of the level to be loaded
     */
    void load_level( const string pathname );

    /**
     * Advance the game's state.  
     *
     * This is where collision detection is done, the ode world stepped, and 
     * the level itself ticked.
     *
     * @param t amount of time (in seconds) to advance the game's state.  note
     *          that ode is very sensitive to this value and it should vary as
     *          little as possible between ticks.
     */
    void tick( double t );

    /**
     * Setup the opengl context and render the current level or a cool
     * background if no level is loaded.
     */
    void render( void );

    /**
     * Reset the currently loaded level.  
     *
     * Currently this amounts to calling load_level().
     *
     * @see GameState::load_level
     */
    void reset_level( void );

    /**
     * Return the opengl state.
     *
     * @return the populated opengl state
     *
     * @see InputHandler::on_key_down
     * @see InputHandler::update_mouse_pos
     * @see InputHandler::on_mouse_down
     * @see InputHandler::on_mouse_move
     */
    CL_OpenGLState* get_glstate( void ) const { return _glstate; }

    //@{
    /** @name Accessors
     *
     * Accessor functions
     */
    Level* get_level( void ) { return _current_level; }

    const State get_current_state( void ) const { return _current_state; }
    const State get_previous_state( void ) const { return _previous_state; }

    void set_current_state( const State state ) { _current_state = state; }
    void set_previous_state( const State state ) { _previous_state = state; }
    //@}

private:

    /**
     * Static wrapper for near_callback.  
     *
     * ode wants a static function passed to dSpaceCollide and that's why this
     * function is necessary.
     *
     * @param data a pointer to the GameState object ('this' pointer)
     * @param o1   the geom id of the first object
     * @param o2   the geom id of the second object
     */
    static void _static_near_callback( void* data, dGeomID o1, dGeomID o2 );

    /**
     * Performs collision detection between two potentially colliding objects.
     *
     * Attaches a contact joint between the two objects if they do collide.
     * 
     * @param o1 the geom id of the first object
     * @param o2 the geom id of the second object
     */
    void _near_callback( dGeomID o1, dGeomID o2 );

    /**
     * Set up the opengl state for level rendering.
     */
    void _gl_init( void );

    Level* _current_level;                  ///< the currently loaded level

    State _current_state, _previous_state;  ///< game state information past and present

    CL_OpenGLState* _glstate;               ///< opengl state information. note: does not store lighting state!
};

#endif // header guard
