/**
 * @file goal.cpp
 * Implementation of Goal class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "goal.h"
#include "object.h"

#include <sstream>

Goal::Goal( int ID, GoalType goal_type, int object_id, Object* object,
            Vec3r position, double radius, double delay ) :
    _id( ID ),
    _goaltype( goal_type ),
    _object_id( object_id ),
    _object( object ),
    _position( position),
    _radius( radius ),
    _delay( delay ),
    _complete_timer( -1.0 )
{

}

Goal::~Goal( void )
{

}

void Goal::tick( const double t )
{
    if ( _complete_timer >= 0.0 ) {
        _complete_timer += t;
    }
}


// may want to support relative positions from the goal object instead of
// just absolute world coordinate positions
const bool Goal::check_goal( void )
{
    bool completed = false;

    if ( !_object ) {
        cerr << "Goal::check_goal: NULL object!" << endl;
        return false;
    }

    switch ( _goaltype ) {
    case INSIDE:

        if ( norm( _object->get_position() - _position ) < _radius ) {
            completed = true;
        }

        break;

    case OUTSIDE:

        if ( norm( _object->get_position() - _position ) > _radius ) {
            completed = true;
        }

        break;

    case ACTIVATE:

        if ( _object->is_on() ) {
            completed = true;
        }

        break;

    case DEACTIVATE:

        if ( !_object->is_on() ) {
            completed = true;
        }

        break;

    default:

        completed = false;

        break;
    }

    if ( completed ) {
        if ( _complete_timer < 0.0 ) {
            // first time.  start the timer.
            _complete_timer = 0.0;
            completed = false;
        }
        else if ( _complete_timer >= _delay ) {
            // it's been long enough, we're completed
            _complete_timer = -1.0;
            completed = true;
        }
        else {
            // nope, still going
            completed = false;
        }
    }
    else {
        _complete_timer = -1.0;
    }

    return completed;
}

string Goal::get_info( void ) const
{
    ostringstream out;

    switch ( _goaltype ) {
    case INSIDE:
        out << "type=\"inside\" ";
        out << "objectID=\"" << _object_id << "\" ";
        out << "position=\"" << _position << "\" ";
        out << "radius=\"" << _radius << "\" ";
        break;
    case OUTSIDE:
        out << "type=\"outside\" ";
        out << "objectID=\"" << _object_id << "\" ";
        out << "position=\"" << _position << "\" ";
        out << "radius=\"" << _radius << "\" ";
        break;
    case ACTIVATE:
        out << "type=\"activate\" ";
        break;
    case DEACTIVATE:
        out << "type=\"deactivate\" ";
        break;
    }

    out << "delay=\"" << _delay << "\" ";

    return out.str();
}
