/**
 * @file goal.h
 * Header file for Goal class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _GOAL_H
#define _GOAL_H

#include "main.h"
#include "gfxutils.h"

#include <string>
using namespace std;

class Object;

/**
 * A single Level objective.
 *
 * A level is considered completed when all its goals are satisfied.
 *
 * @see Level
 */
class Goal {
public:

    /**
     * Different types of goals a level can have.
     */
    enum GoalType {
        INSIDE,         ///< get the ball inside the bucket
        OUTSIDE,        ///< get the mouse out of the cage
        ACTIVATE,       ///< turn on the light bulb
        DEACTIVATE      ///< turn off the fan
    };

    /**
     * Goal constructor.
     *
     * @param ID        a unique identifier for this goal
     * @param goal_type what type of goal this is
     * @param object_id id of the object involved in this goal (necessary for
                        the saving of the level)
     * @param object    pointer to the object involved in this goal
     * @param position  used for INSIDE/OUTSIDE goals, where the object needs
                        to near/away from
     * @param radius    how close/far away the object needs to be to position
                        for the goal to be satisfied
     * @param delay     how long (in seconds) the object needs to be in a
                        satisficing state
     */
    Goal( int ID, GoalType goal_type, int object_id, Object* object,
            Vec3r position = Vec3r( 0.0, 0.0, 0.0 ),
            double radius = 0.0, double delay = 0.0 );

    /**
     * Goal destructor
     *
     * is NOT responsible for deleting its associated object
     */
    ~Goal( void );

    /**
     * Advance the goal's state by updating how long the goal object has
     * satisfied the goal.
     *
     * @param t amount of time (in seconds) to advance the goal's state.
     */
    void tick( const double t );

    /**
     * Check if the goal object has satisfied the goal.
     *
     * @return true if the goal is satisfied, false otherwise
     */
    const bool check_goal( void );

    //@{
    /** @name Accessors
     *
     * Accessor functions.
     */
    Object* get_object( void ) const { return _object; }

    string get_info( void ) const;
    //@}

private:

    int _id;                ///< unique identifier

    GoalType _goaltype;     ///< type of goal

    int _object_id;         ///< unique id of the goal object

    Object* _object;        ///< the goal object

    Vec3r _position;         ///< position of the goal

    double _radius;         ///< sphere surrounding the position

    double _delay;          ///< how long the goal must be satisfied

    double _complete_timer;     ///< how long the goal object has satisfied the goal
};

#endif
