/**
 * @file gamestate/level.cpp
 * Implementation of Gamestate/level class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "level.h"
#include "goal.h"
#include "camera.h"
#include "model.h"
#include "object.h"
#include "seesaw.h"
//#include "effects.h"

#include <fstream>

Level* Level::_current = NULL;

Level::Level( const string pathname ) :
    _pathname( pathname ),
    _max_object_id( 0 ),
    _editor_mode( false ),
    _show_movables( true ),
    _selected_object( NULL ),
    _selected_colliding( false ),
    _axes_type( DRAW_ON_SELECTED )
{
    _current = this;
}

Object *Level::new_object_from_dom( const CL_DomElement& element )
{
    string type = get_string( element, "type" );

    Object::LoadFunc* func = RegisterLoader::get_loader( type );
    if ( func ) {
        return ( *func )( _world_id, _space_id, element );
    }
    else {
        cerr << "Object type not recognized" << endl;
        return NULL;
    }
}

void Level::load( void )
{
    cout << "Loading: " << _pathname << endl;

    _world_id = dWorldCreate();

    dWorldSetGravity( _world_id, 0.0f, -9.81f, 0.0f );
    dWorldSetERP( _world_id, 0.4f );
    dWorldSetCFM( _world_id, 1e-5f );

    _space_id = dSimpleSpaceCreate( 0 );

    _contact_group = dJointGroupCreate( 0 );


    CL_ResourceManager level_resources( _pathname );

    _level_description = CL_String::load( "Description/LevelDescription", &level_resources );
    _select_description = CL_String::load( "Description/SelectDescription", &level_resources );

    _id = CL_Integer( "Description/ID", &level_resources );

    CL_DomElement element;

    CL_Resource camera_resource = level_resources.get_resource( "World/camera" );
    camera_resource.load();
    element = camera_resource.get_element();
    double radius = get_double( element, "radius" );
    double elevation = get_double( element, "elevation" );
    double azimuth = get_double( element, "azimuth" );

    _camera = new Camera( radius, elevation, azimuth );

    CL_Resource world_resource = level_resources.get_resource( "World/world" );
    world_resource.load();
    element = world_resource.get_element();

    Vec3 world_size = get_vec3( element, "size" );
    string wall_texture = get_string( element, "wall" );
    string background_texture = get_string( element, "background" );

    Material wall_mat = grey.copy_with_texture( wall_texture );
    Material background_mat = grey.copy_with_texture( background_texture );

    _world_box = new WorldBox( _space_id, world_size, wall_mat, background_mat );

    list<string>::iterator itr;

    list<string> object_list = level_resources.get_resources_of_type( "object", "Objects" );

    for ( itr = object_list.begin(); itr != object_list.end(); ++itr ) {
        CL_Resource object_resource = level_resources.get_resource( *itr );

        object_resource.load();

        CL_DomElement element = object_resource.get_element();

        int id = CL_String::to_int( element.get_attribute( "ID" ) );

        if( id > _max_object_id )
            _max_object_id = id;

        Object* obj = new_object_from_dom( element );

        if ( obj ) {
            _objects[ id ] = obj;
            obj->set_id( id );
        }
        else {
            cerr << "Error parsing object element" << endl;
        }

        object_resource.unload();
    }


    list<string> goal_list = level_resources.get_resources_of_type( "goal", "Goals" );

    for ( itr = goal_list.begin(); itr != goal_list.end(); ++itr ) {
        CL_Resource goal_resource = level_resources.get_resource( *itr );

        goal_resource.load();

        CL_DomElement element = goal_resource.get_element();

        int id = get_int( element, "ID" );

        string type = get_string( element, "type" );

        int object_id = get_int( element, "objectID" );

        Vec3 position( get_vec3( element, "position" ) );

        double radius = get_double( element, "radius" );

        double delay = get_double( element, "delay" );

        Goal::GoalType goaltype;
        if( CL_String::compare_nocase( type, "inside" ) ) {
            goaltype = Goal::INSIDE;
        }
        else if( CL_String::compare_nocase( type, "outside" ) ) {
            goaltype = Goal::OUTSIDE;
        }
        else if( CL_String::compare_nocase( type, "activate" ) ) {
            goaltype = Goal::ACTIVATE;
        }
        else if( CL_String::compare_nocase( type, "deactivate" ) ) {
            goaltype = Goal::DEACTIVATE;
        }
        else {
            cerr << "Goal not recognized" << endl;
        }

        _goals[ id ] = new Goal( id, goaltype, object_id, _objects[ object_id ], position, radius, delay );

        goal_resource.unload();
    }
#if 0
    // register effect handlers
    RegisterEffect::iterator eit;
    ObjectIterator oit;

    for ( eit = RegisterEffect::begin(); eit != RegisterEffect::end(); eit++ ) {
        (*eit)->clear_handlers();
    }

    for ( oit = _objects.begin(); oit != _objects.end(); oit++ ) {
        for ( eit = RegisterEffect::begin(); eit != RegisterEffect::end(); eit++ ) {
            (*eit)->add_handler( (*oit).second );
        }
    }
#endif
}

Level::~Level( void )
{
    ObjectIterator object_itr;
    GoalIterator goal_itr;

    for( object_itr = _objects.begin(); object_itr != _objects.end(); ++object_itr ) {
        delete object_itr->second;
    }

    for( goal_itr = _goals.begin(); goal_itr != _goals.end(); ++goal_itr ) {
        delete goal_itr->second;
    }

    delete _world_box;

	delete _camera;

    dWorldDestroy( _world_id );
    dSpaceDestroy( _space_id );

    dJointGroupEmpty( _contact_group );
    dJointGroupDestroy( _contact_group );
}

int Level::add_object( string type, string shape )
{
    Object::LoadFunc* func = RegisterLoader::get_loader( type );
    if ( func ) {
        CL_DomDocument doc;
        CL_DomElement element( doc, "object" );
        element.set_attribute( "shape", shape );

        Object* obj = ( *func )( _world_id, _space_id, element );

        if ( obj ) {
            return add_object( obj );
        }
    }
    else {
        cerr << "Unknown Object Type: " << type << endl;
    }

    return -1;
}

int Level::add_object( Object* object )
{
    return add_object( ++_max_object_id, object );
}

int Level::add_object( const int id, Object* object )
{
    if( _objects[ id ] ) {
        return -1;
    }

    if( id > _max_object_id ) {
        _max_object_id = id;
    }

    _objects[ id ] = object;

    return id;
}

void Level::remove_object( Object* object )
{
    ObjectIterator object_itr;

    for ( object_itr = _objects.begin(); object_itr != _objects.end(); ++object_itr ) {
        if ( object_itr->second == object ) {
            delete object;
            _objects.erase( object_itr );
            return;
        }
    }
}

void Level::remove_object( const int id )
{
    remove_object( _objects[ id ] );
}

void Level::tick( const double t )
{
    GoalIterator itr;

    for( itr = _goals.begin(); itr != _goals.end(); ++itr ) {
        itr->second->tick( t );
    }

    ObjectIterator oit;

    // every object should push any non-ode based (magnetism,
    // wind, fear, etc.) effects for this tick into ode
    for ( oit = _objects.begin(); oit != _objects.end(); oit++ ) {
        oit->second->tick( t );
    }
}

void Level::render( bool do_draw_axes )
{
    _camera->apply_view();

    // render the worldbox and background first otherwise it messes
    // with object selection
    _world_box->render( _camera->get_position() );

    Level::ObjectIterator oit;

    for ( oit = _objects.begin(); oit != _objects.end(); oit++ ) {

        Object* object = (*oit).second;

        glLoadName( GLuint( object ) );

        object->render();

        // draw the appropriate bounding box
        if ( ( object == _selected_object ) ) {
            if ( _selected_colliding ) {
                object->draw_decor( Vec3( 1.0, 1.0, 0.0 ) ); // yellow
            }
            if ( object->is_movable() ) {
                object->draw_decor( Vec3( 0.0, 1.0, 0.0 ) ); // green
            }
            else {
                object->draw_decor( Vec3( 1.0, 0.0, 0.0 ) ); // red
            }
        }
        else if ( _show_movables && object->is_movable() ) {
            object->draw_decor( Vec3( 0.0, 0.0, 1.0 ) ); // blue
        }

        if ( do_draw_axes ) {
            draw_axes( object );
        }
    }
}

void Level::draw_axes( Object* object )
{
    double x_r = 0.02, y_r = 0.02, z_r = 0.02;
    double length = 2.0;

    // ARGGGHHH^2, all mouse stuff should be in inputhandler
    switch ( _axes_type ) {
    case DRAW_ON_SELECTED:
        if ( object != _selected_object ) {
            break;
        }
        // fall through
    case DRAW_ON_ALL:
        if ( object == _selected_object ) {
            length *= 2;
            // scale the axes to show which the object moves in
            if ( CL_Mouse::get_keycode( CL_MOUSE_RIGHT ) ) {
                y_r *= 2;
            }
            else {
                x_r *= 2;
                z_r *= 2;
            }
        }

        Model::draw_cylinder( object->get_position( ),
                Vec3( 1, 0, 0 ), red, x_r, length );
        Model::draw_cylinder( object->get_position( ),
                Vec3( 0, 1, 0 ), green, y_r, length );
        Model::draw_cylinder( object->get_position( ),
                Vec3( 0, 0, 1 ), blue, z_r, length );
    default:
        break;
    }
}

const bool Level::level_completed( void )
{
    bool completed = true;

    GoalIterator itr;

    for( itr = _goals.begin(); itr != _goals.end(); ++itr ) {
        if( !itr->second->check_goal() ) {
            completed = false;
            break;
        }
    }

    return completed;
}

void Level::start_sim( void )
{
    ObjectIterator itr;

    // remember the object's current position and orientation so it can be
    // reset after the sim
    for( itr = _objects.begin(); itr != _objects.end(); ++itr ) {
        itr->second->start_sim();
    }
}

void Level::stop_sim( void )
{
    ObjectIterator itr;

    // reset the object's position and orientation to the position and
    // orientation it had before the simulation started
    for( itr = _objects.begin(); itr != _objects.end(); ++itr ) {
        itr->second->stop_sim();
    }
}

void Level::reset_level( void )
{
    // do any other initialization needed
}

void Level::save( const string filename )
{
    if ( filename.empty() ) {
        save( cout );
    }
    else {
        ofstream fout( filename.c_str(), ios::out );
        save( fout );
        fout.close();
    }
}

void Level::select_object( int *pos )
{
    GLuint buffer[ 128 ]; // may need to adjust the size

    // let opengl know we want to pick an object
    begin_opengl_pick( pos, 1.0, buffer, 128 );

    _camera->apply_projection();

    // render the scene so opengl knows where everthing is
    render( false );

    // find out what object was picked
    GLuint select_idx = complete_opengl_pick( buffer );

    if ( select_idx != opengl_pick_nil ) {
        _selected_object = (Object*) select_idx;
    }
}

// return values so the inputhandler knows what to do on return
int Level::deselect_object( void )
{
    if ( !_selected_object ) {
        return NO_SELECTED;
    }

    // the currently selected object is colliding with some other object,
    // don't let the user deselect it in its current position otherwise
    // ode will explode when the simulation is started
    if ( _selected_colliding ) {
        return COLLIDING;
    }

    bool movable = _selected_object->is_movable();

    _selected_object = NULL;

    // may need to update the mouse position after the object is deselected
    if( movable ) {
        return UPDATE_MOUSE;
    }

    return SUCCESS;
}

void Level::_static_near_callback( void* data, dGeomID o1, dGeomID o2 )
{
    ((Level*) data)->_near_callback( o1, o2 );
}

void Level::_near_callback( dGeomID o1, dGeomID o2 )
{
    const unsigned long N = 1;
    dContact contacts[ N ];

    const unsigned long n = dCollide( o1, o2, N, &contacts[ 0 ].geom, sizeof( dContact ) );

    if( n > 0 && ( _selected_object == dBodyGetData( dGeomGetBody( o1 ) ) ||
                   _selected_object == dBodyGetData( dGeomGetBody( o2 ) ) ) ) {

        _selected_colliding = true;
    }
}

void Level::check_selected_colliding( void )
{
    if( !_selected_object ) {
        return;
    }

    _selected_colliding = false;

    // optimization: use dSpaceCollide2() and check if only selected_object 
    // is colliding
    dSpaceCollide( _space_id, this, &Level::_static_near_callback );
}


void Level::save( ostream& out )
{
#define printline( x ) out << spaces << x << endl
#define indent() spaces += "  ";
#define unindent() spaces.erase( 0, 2 );

    string spaces = "";

    printline( "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>" );
    printline( "<resources>" );
    indent();


    printline( "<section name=\"Description\">" );
    indent();

    printline( "<integer name=\"ID\" value=\"" << _id << "\" />" );
    printline( "<string name=\"SelectDescription\" value=\"" << _select_description << "\" />" );
    printline( "<string name=\"LevelDescription\" value=\"" << _level_description << "\" />" );

    unindent();
    printline( "</section>" );
    out << endl;


    printline( "<section name=\"World\">" );
    indent();

    printline( "<world name=\"world\" " << _world_box->get_info() << " />" );
    printline( "<camera name=\"camera\" "
            "radius=\"" << _camera->get_radius() << "\" "
            "elevation=\"" << _camera->get_elevation() << "\" "
            "azimuth=\"" << _camera->get_azimuth() << "\" "
            "/>" );

    unindent();
    printline( "</section>" );
    out << endl;

    printline( "<section name=\"Goals\">" );
    indent();

    for( GoalIterator goal_itr = _goals.begin(); goal_itr != _goals.end(); ++goal_itr ) {
        int goal_id = (*goal_itr).first;
        Goal *goal = (*goal_itr).second;

        printline(
                "<goal "
                "name=\"" << goal_id << "\" "
                "ID=\"" << goal_id << "\" "
                << goal->get_info() <<
                "/>" );
    }

    unindent();
    printline( "</section>" );
    out << endl;

    printline( "<section name=\"Objects\">" );
    indent();

    for( ObjectIterator object_itr = _objects.begin(); object_itr != _objects.end(); ++object_itr ) {
        int obj_id = (*object_itr).first;
        Object* obj = (*object_itr).second;

        printline(
                "<object "
                "name=\"" << obj->get_type_str() << obj_id << "\" "
                "ID=\"" << obj_id << "\" "
                << endl
                << obj->get_info() <<
                "/>" );
    }


    unindent();
    printline( "</section>" );

    unindent();
    printline( "</resources>" );
}
