/**
 * @file level.h
 * Header file for Level class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _LEVEL_H
#define _LEVEL_H

#include "main.h"
#include "worldbox.h"
#include "gfxutils.h"

#include <map>
#include <string>
using namespace std;

#include <ClanLib/core.h>
#include <ClanLib/gl.h>

#include <gfx/gl.h>
#include <gfx/gltools.h>

class Object;
class Goal;
class Camera;

/**
 * A playable level consisting of objects and goals created from an xml level
 * file.
 *
 * Responsible for rendering the level, starting/stopping a simulation of the
 * level and saving the level
 */
class Level {
public:

    /**
     * Different ways the projecting object axes can be drawn
     */
    enum AxesType {
        DRAW_ON_NONE,       ///< don't draw axes
        DRAW_ON_SELECTED,   ///< draw only on selected object
        DRAW_ON_ALL,        ///< draw on all objects
        DRAW_NUM_TYPES      ///< number of types
    };

    //@{
    /** @name Convenience typedef's
     */
    typedef map<int, Object* > ObjectList;
    typedef map<int, Goal* > GoalList;

    typedef ObjectList::iterator ObjectIterator;
    typedef GoalList::iterator GoalIterator;
    //@}

    /**
     * Return the loaded level ('this' pointer)
     *
     * Some object's need to get at information about the level they're a part
     * of while in a function that cannot be passed that information.
     *
     * @see Connectable::load
     */
    static Level* get_current( void ) { return _current; }

    /**
     * Level constructor
     *
     * @param filename relative path + filename of the level to be constructed
     */
    Level( const string pathname = "" );

    /**
     * Level destructor.
     *
     * Delete all objects and goals, and delete the ode world and space
     */
    ~Level( void );

    /**
     * Actually populate the level and its structures.
     *
     * This is not in the constructor because we need Level to be created
     * while loading, so objects can do lookups on other object's IDs.
     */
    void load( void );

    /**
     * Advance the level's state.
     *
     * Currently this ticks every object and goal of the level.
     *
     * @param t amount of time (in seconds) to advance the level's state
     */
    void tick( const double t );

    /**
     * Render every object in the level and the level's worldbox/background.
     *
     * @param draw_axes if true, draw projecting axes, otherwise don't
     */
    void render( bool draw_axes );

    /**
     * Draw projecting axes around an object.
     *
     * @param object the object to draw axes around
     */
    void draw_axes( Object* object );

    /**
     * Start the ode level simulation.
     *
     * Currently just calls start_sim on every object in the level.  Should
     * only be called on the GAMEPLAY->RUN_SIM game state transition.
     */
    void start_sim( void );

    /**
     * Stop the ode level simulation.
     *
     * Currently just calls stop_sim on every object in the level.  Should
     * only be called on the RUN_SIM->GAMEPLAY game state transition.
     */
    void stop_sim( void );

    /**
     * Reset the level.
     *
     * Currently doesn't do a damn thing.
     */
    void reset_level( void );

    /**
     * Is the currently selected object colliding with any other object?
     *
     * Useful for not allowing the user to place an object inside another
     * object and subsequently start the simulation (which makes ode vomit
     * everywhere)
     */
    void check_selected_colliding( void );

    /**
     * Save the level to a file.
     *
     * @param pathname relative path + filename of the file into which to save
     *                 the level
     */
    void save( const string pathname = "" );

    /**
     * Save the level to a stream.
     *
     *  @param out name of the stream into which to save the level
     */
    void save( ostream& out );

    /**
     * Check if the level has been completed.
     *
     * Should be called every main game loop.
     *
     * @return true if the level has been completed, false otherwise
     */
    const bool level_completed( void );

    /**
     * Create an Object from an xml element from the level file.
     *
     * @param element the xml element from the level file consisting of all
     *                the parameters necessary to create the object
     *
     * @return a pointer to the newly created object
     */
    Object *new_object_from_dom( const CL_DomElement& element );

    /**
     * Create an object and add it to the level while the game is running.
     *
     * @param type  the type of object to create
     * @param shape the shape of the object to be created
     *
     * @return the id of the created object on success, -1 on failure
     *
     * @see GUI::create_object
     */
    int add_object( string type, string shape = "" );

    /**
     * Add an object to the level while the game is running.
     *
     * @param object pointer to the object to add
     *
     * @return the id of the added object
     */
    int add_object( Object* object = NULL );

    /**
     * Add an object to the level while the game is running.
     *
     * @param ID     the id of the object to add, must be unique
     * @param object pointer to the object to add
     *
     * @return the id of the added object, should be the same as the id param
     */
    int add_object( const int id = -1, Object* object = NULL );

    /**
     * Remove an object from the level while the game is running.
     *
     * @param object pointer to the object to be removed, delete will be called
                     on this pointer
     */
    void remove_object( Object* object = NULL );

    /**
     * Remove an object from the level while the game is running.
     *
     * @param ID the id of the object to remove, delete will be called on the
     *           corresponding object pointer
     */
    void remove_object( const int id = -1 );

    /**
     * Select the object at screen position pos.
     *
     * _selected_obj will then contain the picked object, or NULL if background
     * was clicked.  Assumes a correct opengl context is in place
     *
     * @param pos     the x, y window coordinates that were selected.
     *
     * @see InputHandler
     */
    void select_object( int* pos );

    /**
     * Attempt to deselect the currently selected object.
     *
     * @return 0 on success
     *         -1 if no object is selected
     *         -2 on success, but the object was movable
     *         -3 if the currently selected object is colliding with some other
     *            object
     *
     * @see InputHandler
     */
    int deselect_object( void );

    //@{
    /** @name Accessors
     *
     * Accessor functions.
     */
    ObjectList& get_objects( void ) { return _objects; }
    GoalList& get_goals( void ) { return _goals; }
    Object* get_selected_object( void ) const { return _selected_object; }
    void set_selected_object( Object* object ) { _selected_object = object; }
    Object* get_object( int id ) { return _objects [ id ]; }
    int get_level_id( void ) const { return _id; }
    string get_pathname( void ) const { return _pathname; }
    string get_level_description( void ) const { return _level_description; }
    string get_select_description( void ) const { return _select_description; }
    Vec3r get_world_size( void ) const { return _world_box->get_size(); }
    Camera* get_camera( void ) { return _camera; }
    dWorldID get_world_id( void ) const { return _world_id; }
    dSpaceID get_space_id( void ) const { return _space_id; }
    dJointGroupID get_contact_group( void ) { return _contact_group; }
    bool is_editor_mode( void ) const { return _editor_mode; }
    void set_axes_type( AxesType axes_type ) { _axes_type = axes_type; }
    AxesType get_axes_type( void ) const { return _axes_type; }
    //@}

    void toggle_show_movables( void ) { _show_movables = !_show_movables; }
    void toggle_editor_mode( void ) { _editor_mode = !_editor_mode; }

private:

    /**
     * Static wrapper for near_callback.
     *
     * ode wants a static function passed to dSpaceCollide and that's why this
     * function is necessary.
     *
     * @param data a pointer to the Level object ('this' pointer)
     * @param o1   the geom id of the first object
     * @param o2   the geom id of the second object
     */
    static void _static_near_callback( void* data, dGeomID o1, dGeomID o2 );

    /**
     * Performs collision detection between two potentially colliding objects.
     *
     * Does not create contact joints between the two objects if they are
     * colliding, just used for checking if the currently selected object is
     * colliding with some other object.
     *
     * @param o1   the geom id of the first object
     * @param o2   the geom id of the second object
     */
    void _near_callback( dGeomID o1, dGeomID o2 );

    int _id;                        ///< unique identifier

    int _max_object_id;             ///< running max of the highest object id in the level

    string _pathname;               ///< relative path + filename of the level

    string _select_description;     ///< description used in level selection box
    string _level_description;      ///< description used in level loading box

    ObjectList _objects;            ///< all the objects in the level
    GoalList _goals;                ///< all the goals of the level


    Object* _selected_object;       ///< currently selected object

    bool _editor_mode;              ///< allow editor commands

    bool _show_movables;            ///< show movable objects in blue

    bool _selected_colliding;       ///< is the selected object colliding with another object


    WorldBox* _world_box;

    Camera* _camera;


    dWorldID _world_id;             ///< ode world in which to place object bodies
    dSpaceID _space_id;             ///< ode space in which to place object geometries
    dJointGroupID _contact_group;   ///< joint group in which to place contact joints resulting from collision detection


    AxesType _axes_type;            ///< in which way to draw projecting axes

    static Level* _current;         /// pointer to the current level ('this' pointer)
};

#endif // header guard
