/**
 * @file gui/gui.cpp
 * Implementation of gui/gui class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "gui.h"
#include "gamestate.h"
#include "level.h"
#include "model.h"
#include "shapemodel.h"
#include "msmodel.h"
#include "object.h"
#include "shapeobject.h"
#include "bucket.h"
#include "seesaw.h"
#include "conveyorbelt.h"
#include "trigger.h"

GUI::GUI( void ) :
    _window( NULL ),
    _gui_manager( NULL ),
    _style( NULL ),
    _component_manager( NULL ),
    _resource_manager( NULL ),
    _gamestate( NULL ),
    _quit( false ),
    _ignore_move_signal( false ),
    _window_moved( false ),
    _show_esc_dialog( false )
{
    // there are weird seg-fault issues with _window on my MSVC6 + XP box,
    // tracked down to the EnumDevices call in CL_DisplayWindow_Win32::create_new_window


    // the first member of the struct needs to match the name value of the
    // window in the xml file while the second member needs to be the
    // corresponding game_state.  To be able to use game_state->current_game_state
    // as an index into this array THE MENU ORDER MUST MATCH THE ORDER OF
    // THE STATE ENUM IN GAMESTATE.H
    _menu_names.push_back( MenuName( "Main",                GameState::MAIN                ) );
    _menu_names.push_back( MenuName( "Level Loading",       GameState::LEVEL_LOADING       ) );
    _menu_names.push_back( MenuName( "GamePlay",            GameState::GAMEPLAY            ) );
    _menu_names.push_back( MenuName( "End Level",           GameState::END_LEVEL           ) );
    _menu_names.push_back( MenuName( "End Game",            GameState::END_GAME            ) );
    _menu_names.push_back( MenuName( "Level Select",        GameState::LEVEL_SELECT        ) );
    _menu_names.push_back( MenuName( "Ars Physica Help",    GameState::HELP                ) );
    _menu_names.push_back( MenuName( "Options Video Sound", GameState::OPTIONS_VIDEO_SOUND ) );
    _menu_names.push_back( MenuName( "Options Controls",    GameState::OPTIONS_CONTROLS    ) );
    _menu_names.push_back( MenuName( "Esc Dialog",          GameState::ESC_DIALOG          ) );
    _menu_names.push_back( MenuName( "Run Sim",             GameState::RUN_SIM             ) );

    // the first member of the struct needs to match the name value of the
    // various buttons in the xml file while the second member needs to be
    // the corresponding handler
    _button_handlers.push_back( ButtonHandler( "StartGame",       &GUI::_on_start_game      ) );
    _button_handlers.push_back( ButtonHandler( "SelectLevel",     &GUI::_on_select_level    ) );
    _button_handlers.push_back( ButtonHandler( "Options",         &GUI::_on_options         ) );
    _button_handlers.push_back( ButtonHandler( "StartLevel",      &GUI::_on_start_level     ) );
    _button_handlers.push_back( ButtonHandler( "Help",            &GUI::_on_help            ) );
    _button_handlers.push_back( ButtonHandler( "SelectCancel",    &GUI::_on_cancel          ) );
    _button_handlers.push_back( ButtonHandler( "LevelLoadOK",     &GUI::_on_level_load      ) );
    _button_handlers.push_back( ButtonHandler( "HelpOK",          &GUI::_on_help_ok         ) );
    _button_handlers.push_back( ButtonHandler( "MainQuit",        &GUI::_on_quit            ) );
    _button_handlers.push_back( ButtonHandler( "EscNewLevel",     &GUI::_on_select_level    ) );
    _button_handlers.push_back( ButtonHandler( "SetOptions",      &GUI::_on_options_ok      ) );
    _button_handlers.push_back( ButtonHandler( "ControlOptions",  &GUI::_on_control_options ) );
    _button_handlers.push_back( ButtonHandler( "SetControls",     &GUI::_on_set_controls    ) );
    _button_handlers.push_back( ButtonHandler( "StartSim",        &GUI::_on_start_sim       ) );
    _button_handlers.push_back( ButtonHandler( "EscHelp",         &GUI::_on_help            ) );
    _button_handlers.push_back( ButtonHandler( "EscMainMenu",     &GUI::_on_esc_main_menu   ) );
    _button_handlers.push_back( ButtonHandler( "EscResetLevel",   &GUI::_on_reset_level     ) );
    _button_handlers.push_back( ButtonHandler( "EscLevelInfo",    &GUI::_on_level_info      ) );
    _button_handlers.push_back( ButtonHandler( "EscQuit",         &GUI::_on_quit            ) );
    _button_handlers.push_back( ButtonHandler( "NextLevel",       &GUI::_on_next_level      ) );
    _button_handlers.push_back( ButtonHandler( "Solution",        &GUI::_on_solution        ) );
    _button_handlers.push_back( ButtonHandler( "ResetSim",        &GUI::_on_stop_sim        ) );
    _button_handlers.push_back( ButtonHandler( "EscSave",         &GUI::_on_save            ) );
    _button_handlers.push_back( ButtonHandler( "EscHints",        &GUI::_on_hints           ) );

    // objects that can be created with the in-game editor
    _creatable_objects[ "Shape Object - Box" ]      = CreatableObject( "Shape Object - Box",      "shape", "box" );
    _creatable_objects[ "Shape Object - Sphere" ]   = CreatableObject( "Shape Object - Sphere",   "shape", "sphere" );
    _creatable_objects[ "Shape Object - Cylinder" ] = CreatableObject( "Shape Object - Cylinder", "shape", "cylinder" );
    _creatable_objects[ "Seesaw" ]                  = CreatableObject( "Seesaw",                  "seesaw", "" );
    _creatable_objects[ "Conveyor Belt" ]           = CreatableObject( "Conveyor Belt",           "conveyorbelt", "" );
    _creatable_objects[ "Bucket" ]                  = CreatableObject( "Bucket",                  "bucket", "" );
    _creatable_objects[ "Trigger - Box" ]           = CreatableObject( "Trigger - Box",           "trigger", "box" );
    _creatable_objects[ "Trigger - Sphere" ]        = CreatableObject( "Trigger - Sphere",        "trigger", "sphere" );
    _creatable_objects[ "Trigger - Cylinder" ]      = CreatableObject( "Trigger - Cylinder",      "trigger", "cylinder" );
}

GUI::~GUI( void )
{
	if( _component_manager ) {
		delete _component_manager;
	}

	if( _gui_manager ) {
		delete _gui_manager;
	}

	if( _style ) {
		delete _style;
	}

	if( _resource_manager ) {
		delete _resource_manager;
	}

    // HACK: this segfaults in Linux.  wait till it gets fixed.
    // delete _window;
}

int GUI::add_resources( const string resource_name )
{
    if ( !_resource_manager ) {
        if( ( _resource_manager = new CL_ResourceManager( resource_name ) ) == NULL ) {
            return -1;
        }
    }
    else {
        _resource_manager->add_resources( CL_ResourceManager( resource_name ) );
    }

    return 0;
}

int GUI::initialize( void )
{
    if ( !_resource_manager )
        return -1;

    string title = get_string( get_element( *_resource_manager, "title" ), "value" );
    int width = get_int( get_element( *_resource_manager, "width" ), "value" );
    int height = get_int( get_element( *_resource_manager, "height" ), "value" );
    bool fullscreen = get_bool( get_element( *_resource_manager, "fullscreen" ), "value" );

    if ( ( _window = new CL_DisplayWindow( title, width, height, fullscreen ) ) == NULL )
        return -1;

    if ( ( _style = new CL_StyleManager_Silver( _resource_manager ) ) == NULL )
        return -1;

    if ( ( _gui_manager = new CL_GUIManager( _style ) ) == NULL )
        return -1;

    if ( ( _component_manager = new CL_ComponentManager( "gui", _resource_manager, _gui_manager ) ) == NULL )
        return -1;

    // context window and edit window should probably be defined in xml

    _context_window = new CL_Window( CL_Rect( 10, 10, 300, 200 ), "Create Object Menu", _gui_manager );

    _context_window->show( false );

    _edit_window = new CL_Window( CL_Rect( 10, 10, 300, 200 ), "Edit Object Menu", _gui_manager );

    _edit_window->show( false );


    _on_check_box = new CL_CheckBox( CL_Point( 10, 25 ), "On?", _edit_window );
    _movable_check_box = new CL_CheckBox( CL_Point( 10, 40 ), "Movable?", _edit_window );
    _stationary_check_box = new CL_CheckBox( CL_Point( 10, 55 ), "Stationary?", _edit_window );

    CL_Label* scale_label = new CL_Label( CL_Point( 10, 75 ), "Scale", _edit_window );

    _scale_scroll_bar = new CL_ScrollBar( CL_Rect( 10, 90, 160, 105 ), -255, 255, true, _edit_window );
    _scale_scroll_bar->set_value( 0 );

    CL_Label* mass_label = new CL_Label( CL_Point( 10, 115 ), "Mass", _edit_window );
    _mass_input_box = new CL_InputBox( CL_Rect( 40, 110, 100, 130 ), _edit_window );

    CL_Label* rotatable_label = new CL_Label( CL_Point( 10, 135 ), "Rotable", _edit_window );
    _rotX_check_box = new CL_CheckBox( CL_Point( 55, 135 ), "X", _edit_window );
    _rotY_check_box = new CL_CheckBox( CL_Point( 85, 135 ), "Y", _edit_window );
    _rotZ_check_box = new CL_CheckBox( CL_Point( 115, 135 ), "Z", _edit_window );

    _edit_ok_button = new CL_Button( CL_Point( 125, 160 ), "OK", _edit_window );

    _slot_edit_ok = _edit_ok_button->sig_clicked().connect( this, &GUI::_on_edit_ok );


    _context_list_box = new CL_ListBox( CL_Rect( 10, 30, 200, 175 ), _context_window );

    _context_list_box->clear();

    map<string, CreatableObject>::iterator itr;

    for( itr = _creatable_objects.begin(); itr != _creatable_objects.end(); itr++ ) {
        _context_list_box->insert_item( itr->first );
    }

    _create_ok_button = new CL_Button( CL_Point( 225, 75 ), "OK", _context_window );

    _slot_create_ok = _create_ok_button->sig_clicked().connect( this, &GUI::_create_object );


    float scale_x = CL_Display::get_width()  / 800.0,
          scale_y = CL_Display::get_height() / 600.0;

    for ( uint i = 0; i < _menu_names.size(); i++ ) {
        CL_Window* window = (CL_Window *) _get_component( _menu_names[ i ].name );

        // by default, no menus should be visible
        window->show( false );

        // scale the size and position of the window appropriately
        window->set_size( int( scale_x * window->get_width() ),
                          int( scale_y * window->get_height() ) );

        window->set_position( int( scale_x * window->get_position().left ),
                              int( scale_y * window->get_position().top ) );

        // get the original position of this window so we can move it back if the
        // user tries to manually move it
        _window_rects.push_back( window->get_position() );
    }

    // scale the buttons
    for ( uint k = 0; k < _button_handlers.size(); k++ ) {
        CL_Button* button = (CL_Button *) _get_component( _button_handlers[ k ].name );

        button->set_size( int( scale_x * button->get_width() ),
                          int( scale_y * button->get_height() ) );

        button->set_position( int( scale_x * button->get_position().left ),
                              int( scale_y * button->get_position().top ) );
    }

    // scale the labels
    char* label_names[] = { "LevelLoadingText", "LevelEndText", "HelpText" };
    for ( int m = 0; m < 3; m++ ) {
        CL_Label* label = (CL_Label *) _get_component( label_names[ m ] );

        label->set_size( int( scale_x * label->get_width() ),
                         int( scale_y * label->get_height() ) );

        label->set_position( int( scale_x * label->get_position().left ),
                             int( scale_y * label->get_position().top ) );
    }

    // scale the listbox
    CL_ListBox* list_box = (CL_ListBox *) _get_component( "LevelList" );

    list_box->set_size( int( scale_x * list_box->get_width() ),
                        int( scale_y * list_box->get_height() ) );

    list_box->set_position( int( scale_x * list_box->get_position().left ),
                            int( scale_y * list_box->get_position().top ) );


    // NOTE: we aren't scaling the font size, so the buttons and menus look
    // larger for larger resolutions but the font stays small


    // special handling for GameState::GAMEPLAY and GameState::RUN_SIM states since they're made up
    // of buttons and not windows
    ((CL_Button *) _get_component( "StartSim" ) )->show( false );
    ((CL_Button *) _get_component( "ResetSim" ) )->show( false );


    // get all the level file info and populate the level select list box
    ((CL_ListBox *) _get_component( "LevelList" ) )->clear();

    CL_DirectoryScanner scanner;

    // find all files of the form "level-*.xml" (case-insensitive)
    if ( scanner.scan( "resources/levels/", "level-*.xml" ) )
    {
        while ( scanner.next() )
        {
            string filename = scanner.get_pathname();
            string solution_filename( filename );

            // construct the solution's file name
            string::size_type pos = solution_filename.rfind( "level-" );
            if ( pos != string::npos ) {
                solution_filename.replace( pos, strlen( "level-" ), "solution-" );
            }

            CL_ResourceManager level_resources( filename );

            string description = CL_String::load( "Description/SelectDescription", &level_resources );

            _level_info[ description ] = LevelInfo( filename, solution_filename, description );

            ((CL_ListBox *) _get_component( "LevelList" ) )->insert_item( description );
        }
    }

    _update_menus();

    return 0;
}

int GUI::register_signals( void )
{
    if ( !_gui_manager || !_window )
        return -1;

    // register every button on every button with their respective handlers
    for ( uint i = 0; i < _button_handlers.size(); i++ )
    {
        _slot_button.push_back( ( (CL_Button *) _get_component( _button_handlers[ i ].name ) )->sig_clicked().connect(
                this, _button_handlers[ i ].handler ) );
    }

    // register all the window close and move signals with two pseudo-dummy handlers
    for ( uint j = 0; j < _menu_names.size(); j++ )
    {
        _slot_gui_window_close.push_back( ( (CL_Window *) _get_component( _menu_names[ j ].name ) )->sig_close().connect_virtual(
                this, &GUI::_on_gui_window_close ) );

        _slot_gui_window_move.push_back( ( (CL_Window *) _get_component( _menu_names[ j ].name ) )->sig_move().connect(
                this, &GUI::_on_gui_window_move ) );
    }

    // do any "beginning of frame" processing
    _slot_paint = _gui_manager->sig_paint().connect( this, &GUI::_on_paint );

    // only allow the application to be closed by closing the actual OS window
    _slot_quit = _window->sig_window_close().connect( this, &GUI::_on_main_window_close );

    return 0;
}

CL_Component* GUI::_get_component( string component )
{
    if( !_component_manager ) {
        cerr << "GUI::_get_component: _component_manager == NULL!" << endl;
        return NULL;
    }
    else {
        return _component_manager->get_component( component );
    }
}

void GUI::_on_paint( void )
{
    // CL_Display::clear( CL_Color( 150, 150, 150 ) );
}

void GUI::_on_main_window_close( void )
{
    _quit = true;
}

void GUI::toggle_esc_dialog( void )
{
    if ( !_gamestate )
        return;

    _show_esc_dialog = !_show_esc_dialog;

    // toggle between the two game states
    _gamestate->change_state( _show_esc_dialog ? GameState::ESC_DIALOG : GameState::GAMEPLAY );

    _update_menus();
}

void GUI::_on_gui_window_close( CL_SlotParent_v0 &parent_handler )
{
    // use connect virtual so this message isnt passed any higher, and since
    // this function effectively does nothing the effect is that GUI windows
    // can't be manually closed
}

void GUI::_on_gui_window_move( int x, int y )
{
    if ( _ignore_move_signal )
        return;

    // can't do the actual positioning back to the original location in this
    // function otherwise we enter an infinite loop and crash n' burn
    _window_moved = true;
}

void GUI::_on_start_game( void )
{
    if ( !_gamestate )
        return;

    cout << "Starting Game..." << endl;

    // move actual filename out to a #define or static or global resource file
    _gamestate->load_level( "resources/levels/level-1.xml" );

    ((CL_Label *) _get_component( "LevelLoadingText" ) )->set_text( _gamestate->get_level()->get_level_description() );

    _gamestate->change_state( GameState::LEVEL_LOADING );

    _update_menus();
}

void GUI::_on_select_level( void )
{
    if ( !( _gamestate == NULL ) )
        return;

    cout << "Selecting Level..." << endl;

    _gamestate->change_state( GameState::LEVEL_SELECT );

    _update_menus();
}

void GUI::_on_start_level( void )
{
    if ( !( _gamestate == NULL ) )
        return;

    // nothing's been selected
    if ( ((CL_ListBox *) _get_component( "LevelList" ) )->get_current_item() == -1 )
        return;

    string selected_level = _level_info[ ((CL_ListBox *) _get_component( "LevelList" ) )->get_current_text() ].pathname;

    cout << "Selected Level: " << selected_level << endl;

    _gamestate->load_level( selected_level );

    ((CL_Label *) _get_component( "LevelLoadingText" ) )->set_text( _gamestate->get_level()->get_level_description() );

    _gamestate->change_state( GameState::LEVEL_LOADING );

    _update_menus();
}

void GUI::_on_help( void )
{
    if ( !_gamestate )
        return;

    cout << "Displaying Help..." << endl;

    _gamestate->change_state( GameState::HELP );

    _update_menus();
}

void GUI::_on_cancel( void )
{
    if ( !_gamestate )
        return;

    cout << "Returning to previous state..." << endl;

    _gamestate->change_state( _gamestate->get_previous_state() );

    _update_menus();
}

void GUI::_on_level_load( void )
{
    if ( !_gamestate )
        return;

    cout << "Starting Level..." << endl;

    _gamestate->change_state( GameState::GAMEPLAY );

    _update_menus();
}

void GUI::_on_help_ok( void )
{
    if ( !_gamestate )
        return;

    cout << "Returning to previous state..." << endl;

    _gamestate->change_state( _gamestate->get_previous_state() );

    _update_menus();
}

void GUI::_on_quit( void )
{
    cout << "Quitting..." << endl;

    _quit = true;
}

void GUI::_on_save( void )
{
    if ( !_gamestate || !_gamestate->get_level() )
        return;

    cout << "Saving..." << endl;
    cout << endl;

    _gamestate->get_level()->save( "save.xml" );

    cout << endl << "Done." << endl;
}

void GUI::_on_esc_main_menu( void )
{
    if ( !_gamestate )
        return;

    cout << "Returning to Main menu..." << endl;

    toggle_esc_dialog();

    _gamestate->change_state( GameState::MAIN );

    _update_menus();
}

void GUI::_on_change_options( void )
{
    if ( !_gamestate )
        return;

    cout << "Changing Options..." << endl;

    toggle_esc_dialog();

    _gamestate->change_state( GameState::OPTIONS_VIDEO_SOUND );

    _update_menus();
}

void GUI::_on_options( void )
{
    if ( !_gamestate )
        return;

    cout << "Changing Options..." << endl;

    _gamestate->change_state( GameState::OPTIONS_VIDEO_SOUND );

    _update_menus();
}

void GUI::_on_options_ok( void )
{
    if ( !_gamestate )
        return;

    cout << "Setting Options..." << endl;

    // set all selected options here

    if ( _gamestate->get_previous_state() == GameState::GAMEPLAY ) {
        _gamestate->change_state( GameState::GAMEPLAY );
    }
    else {
        _gamestate->change_state( GameState::MAIN );
    }

    _update_menus();
}

void GUI::_on_control_options( void )
{
    if ( !_gamestate )
        return;

    cout << "Changing Control Options..." << endl;

    // together with the next function, yet another small hack so that we
    // can keep track of where we came from in this doubly-nested menu
    GameState::State temp = _gamestate->get_previous_state();

    _gamestate->change_state( GameState::OPTIONS_CONTROLS );

    _update_menus();

    _gamestate->set_previous_state( temp );
}

void GUI::_on_set_controls( void )
{
    if ( !_gamestate )
        return;

    cout << "Setting Control Options..." << endl;

    // set control options here

    GameState::State temp = _gamestate->get_previous_state();

    _gamestate->change_state( GameState::OPTIONS_VIDEO_SOUND );

    _update_menus();

    _gamestate->set_previous_state( temp );
}

void GUI::_on_start_sim( void )
{
    if ( !_gamestate )
        return;

    _gamestate->change_state( GameState::RUN_SIM );

    _update_menus();
}

void GUI::_on_stop_sim( void )
{
    if ( !_gamestate )
        return;

    _gamestate->change_state( GameState::GAMEPLAY );

    _update_menus();
}

void GUI::_on_reset_level( void )
{
    if ( !_gamestate )
        return;

    _gamestate->load_level( _level_info[ _gamestate->get_level()->get_select_description() ].pathname );

    _gamestate->change_state( GameState::GAMEPLAY );

    _update_menus();
}

void GUI::_on_next_level( void )
{
    if ( !_gamestate )
        return;

    string next_level( "resources/levels/level-1.xml" );

    // populate next_level with name of actual next level

    _gamestate->load_level( next_level );

    _gamestate->change_state( GameState::LEVEL_LOADING );

    _update_menus();
}

void GUI::_on_solution( void )
{
    if ( !_gamestate )
        return;

    _gamestate->load_level( _level_info[ _gamestate->get_level()->get_select_description() ].solution_pathname );

    _gamestate->change_state( GameState::LEVEL_LOADING );

    _update_menus();
}

void GUI::_on_level_info( void )
{
    if ( !_gamestate )
        return;

    _gamestate->change_state( GameState::LEVEL_LOADING );

    _update_menus();
}

void GUI::_on_hints( void )
{

}

void GUI::_create_object( void )
{
    if( !_gamestate || !_gamestate->get_level() ) {
        cerr << "GUI::_create_object: creating object in NULL level!" << endl;
        return;
    }

    _gamestate->get_level()->add_object( _creatable_objects[ _context_list_box->get_current_text() ].type,
                                         _creatable_objects[ _context_list_box->get_current_text() ].shape );

    _context_window->show( false );
}

void GUI::edit_object( Object* object )
{
    if( !object || !_gamestate || !_gamestate->get_level() ) {
        cerr << "GUI::edit_object: NULL parameters!" << endl;
        return;
    }

    _gamestate->get_level()->deselect_object();
    _edited_object = object;

    _on_check_box->set_checked( object->is_on() );

    _movable_check_box->set_checked( object->is_movable() );
    _stationary_check_box->set_checked( object->is_stationary() );

    _mass_input_box->set_text( CL_String::from_float( object->get_mass() ) );

    _rotX_check_box->set_checked( object->is_rotatable( 0 ) );
    _rotY_check_box->set_checked( object->is_rotatable( 1 ) );
    _rotZ_check_box->set_checked( object->is_rotatable( 2 ) );

    _edit_window->show( true );
}

void GUI::_on_edit_ok( void )
{
    if( !_edited_object ) {
        return;
    }

    _edited_object->set_on( _on_check_box->is_checked() );
    _edited_object->set_movable( _movable_check_box->is_checked() );
    _edited_object->set_stationary( _stationary_check_box->is_checked() );


    bool rotatable[ 3 ] = { _rotX_check_box->is_checked(),
                            _rotY_check_box->is_checked(),
                            _rotZ_check_box->is_checked() };

    _edited_object->set_rotatable( rotatable );

    int scale = _scale_scroll_bar->get_value();

    float new_scale = (float) scale / _scale_scroll_bar->get_min_value();

    if ( scale > 0 ) {
        new_scale = (float) scale / _scale_scroll_bar->get_min_value();
    }
    else if ( scale < 0 ) {
        new_scale = 10.0 * (float) scale / _scale_scroll_bar->get_max_value();
    }
    else {
        new_scale = 1.0;
    }

    // _edited_object->set_scale( Vec3( new_scale, new_scale, new_scale ) );

    _edit_window->show( false );

}

void GUI::_update_menus( void )
{
    if ( !_gamestate )
        return;

    GameState::State current_state = _gamestate->get_current_state();
    GameState::State previous_state = _gamestate->get_previous_state();

    // hide the previous menu and show the current one
    ((CL_Window *) _get_component( _menu_names[ previous_state ].name ) )->show( false );
    ((CL_Window *) _get_component( _menu_names[ current_state ].name ) )->show( true );

    // special handling for GameState::GAMEPLAY/RUN_SIM GameState (windows shouldn't be visible
    // and actual GUI elements are just buttons)
    if ( current_state == GameState::GAMEPLAY || current_state == GameState::RUN_SIM ) {
        ((CL_Window *) _get_component( _menu_names[ current_state ].name ) )->show( false );

        ((CL_Button *) _get_component( "StartSim" ) )->show(  (current_state == GameState::GAMEPLAY) );
        ((CL_Button *) _get_component( "ResetSim" ) )->show( !(current_state == GameState::GAMEPLAY) );
    }
    else if ( previous_state == GameState::GAMEPLAY || previous_state == GameState::RUN_SIM ) {
        ((CL_Button *) _get_component( "StartSim" ) )->show( false );
        ((CL_Button *) _get_component( "ResetSim" ) )->show( false );
    }
}

int GUI::render( void )
{
    if ( !_gui_manager || !_gamestate )
        return -1;

    // more hack-ish code because I don't know how to set the windows to be
    // stationary in a more elegant fashion
    if ( _window_moved ) {
        GameState::State current_state = _gamestate->get_current_state();

        // use ignore_move_signal to keep us out of an infinite loop
        // NOTE: could also do this by disconnecting the signal and reconnecting
        _ignore_move_signal = true;

        ((CL_Window *) _get_component( _menu_names[ current_state ].name ) )->set_position( _window_rects[ current_state ].left, _window_rects[ current_state ].top );

        _ignore_move_signal = false;
        _window_moved = false;
    }

    // get changes made by other parts of the system since the last frame
    _update_menus();

    // HACK: clanlib should handle this part
    glPushAttrib( GL_LIGHTING_BIT );
    glDisable( GL_LIGHTING );

    _gui_manager->show();

    glPopAttrib();

    return 0;
}

CL_GraphicContext* GUI::get_gc( void ) const
{
    if ( _window == NULL )
        return NULL;

    return _window->get_gc();
}

const bool GUI::quit( void )
{
    return _quit;
}
