/**
 * @file gui.h
 * Header file for GUI class
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _GUI_H
#define _GUI_H

#include "main.h"
#include "utils.h"
#include "gamestate.h"

#include <string>
#include <map>
#include <vector>
using namespace std;

#include <ClanLib/core.h>
#include <ClanLib/gui.h>
#include <ClanLib/gl.h>
#include <ClanLib/guistylesilver.h>
#include <ClanLib/application.h>
#include <ClanLib/display.h>

class GUI;
class Object;
class GameState;

/**
 * typedef just to get clanlib to stop complaining
 *
 * @see ButtonHandler
 */
typedef void (GUI::*callback)();

/**
 * Specifies the name of a gui button along with its associated handler
 */
struct ButtonHandler {
    string name;
    callback handler;

    ButtonHandler( string name, callback handler ) :
        name( name ), handler( handler ) { }
};

/**
 * Specifies the name of a gui menu along with its associated handler
 */
struct MenuName {
    string name;
    GameState::State game_state;

    MenuName( string name, GameState::State game_state ) :
        name( name ), game_state( game_state ) { }
};

/**
 * Specification of an object that can be created by the level editor
 */
struct CreatableObject {
    string description;
    string type;
    string shape;

    CreatableObject( void ) :
    description( "" ), type( "" ), shape( "" ) { }

    CreatableObject( string description, string type, string shape ) :
        description( description ), type( type ), shape( shape ) { }
};

/**
 * Information about each potentially loadable level
 */
struct LevelInfo {
    string pathname;		    ///< "resources/levels/level-1.xml"
    string solution_pathname;	///< "resources/levels/solution-1.xml"
    string description;         ///< "Level 1 - A Bucket Too Far"

    /**
     * Default level constructor.
     *
     * Necessary to be able to add LevelInfo's to stl maps
     */
    LevelInfo( void ) :
        pathname( "" ), solution_pathname( "" ), description( "" ) { }

    LevelInfo( string pathname, string solution_pathname, string description ) :
        pathname( pathname ),
        solution_pathname( solution_pathname ),
        description( description ) { }
};

/**
 * The graphical user interface of the game.  
 * 
 * Responsible for drawing menus, buttons and the like and taking action based
 * on input to those components.
 *
 * @see InputHandler
 */
class GUI {
public:

    /**
     * Default constructor.
     *
     * Fills in menu and button handlers but does NOT create any gui elements
     */
    GUI( void );

    ~GUI( void );

    /**
     * Add a resource to the GUI.  
     *
     * The two currenly important resources are the xml style file and the xml 
     * window/button file.  All resources must be loaded before initialize 
     * is called!
     *
     * @param resource_name path + filename of the resource to be loaded
     */
    int add_resources( const string resource_name );

    /**
     * Initialize the gui and its resources.  
     *
     * Creates the window, gui manager, style manager and compoment manager and
     * populates the LevelInfo vector. Must be called before register_signals!
     *
     * @return 0 on success, -1 on failure
     */
    int initialize( void );

    /**
     * Register the button and window signal handlers.
     *
     * @return 0 on success, -1 on failure
     */
    int register_signals( void );

    /**
     * Update the gui's state and render the current menu.
     *
     * Responsible for moving moved windows back to their original positions
     *
     * @return 0 on success, -1 on failure
     */
    int render( void );

    CL_GraphicContext* get_gc( void ) const;

    const bool quit( void );

    /**
     * Restore the screen to its normal resolution.
     *
     * Useful for when the game is run in fullscreen mode
     */
    void finish( void ) { _window->set_windowed(); }

    /**
     * Return all the button handlers
     *
     * @see InputHandler::on_mouse_down
     */
    vector<ButtonHandler>& get_button_handlers( void ) { return _button_handlers; }

    CL_Button* get_button( string name ) const { return (CL_Button*) _component_manager->get_component( name ); }

    void toggle_esc_dialog( void );

    void show_context_window( void ) const { _context_window->show( true ); }

    /**
     * Display the editor menu for the currenly selected object.
     *
     * @param object the currently selected object
     */
    void edit_object( Object* object );

    void set_gamestate( GameState* gamestate ) { _gamestate = gamestate; }

private:

    /**
     * Update which menus/buttons are shown/hidden based on what state the game
     * is currently in
     */
    void _update_menus( void );

    //@{
    /** @name Window Handlers
     *
     * Signal handlers for window-related signals
     */
    void _on_paint( void );
    void _on_main_window_close( void );
    void _on_gui_window_close( CL_SlotParent_v0 &parent_handler );
    void _on_gui_window_move( int x, int y );
    //@}

    //@{
    /** @name Button Handlers
     *
     * Signal handlers for button-related signals
     */
    void _on_start_game( void );
    void _on_select_level( void );
    void _on_start_level( void );
    void _on_help( void );
    void _on_cancel( void );
    void _on_level_load( void );
    void _on_help_ok( void );
    void _on_quit( void );
    void _on_change_options( void );
    void _on_options( void );
    void _on_options_ok( void );
    void _on_control_options( void );
    void _on_set_controls( void );
    void _on_esc_main_menu( void );
    void _on_start_sim( void );
    void _on_stop_sim( void );
    void _on_reset_level( void );
    void _on_next_level( void );
    void _on_solution( void );
    void _on_level_info( void );
    void _on_save( void );
    void _on_hints( void );
    void _on_edit_ok( void );
    //@}

    CL_Component* _get_component( string component );
    
    /**
     * Place an object into the level.
     *
     * The type of the object is determine by the currently selected object
     * type/shape in the context menu.
     */
    void _create_object( void );

    CL_DisplayWindow* _window;                      ///< the top-level window

    CL_GUIManager* _gui_manager;                    ///< manages the GUI hierarchy

    CL_StyleManager_Silver* _style;                 ///< manages the style (GUIStyleSilver)

    CL_ComponentManager* _component_manager;        //< manages components (windows/buttons)

    CL_ResourceManager* _resource_manager;          ///< manages resource (xml) files

    vector<CL_Slot > _slot_button;

    vector<CL_Slot > _slot_gui_window_close;
    vector<CL_Slot > _slot_gui_window_move;

    vector<CL_Rect > _window_rects;

    CL_Slot _slot_paint;

    CL_Slot _slot_quit;


    GameState* _gamestate;      ///< need to be able to know which state the game is in

    bool _quit;

    bool _window_moved, _ignore_move_signal;    ///< to get around the infinite loop caused by moving a window in the window_move handler

    bool _show_esc_dialog;


    map<string, LevelInfo > _level_info;

    vector<ButtonHandler > _button_handlers;

    vector<MenuName > _menu_names;


    //@{
    /** @name Level Editor
     *
     * Level editor related variables.
     */
    CL_Window* _context_window;
    CL_Window* _edit_window;

    CL_Button* _create_ok_button;
    CL_Slot _slot_create_ok;

    CL_Button* _edit_ok_button;
    CL_Slot _slot_edit_ok;

    CL_ListBox* _context_list_box;

    CL_CheckBox* _on_check_box ;
    CL_CheckBox* _movable_check_box;
    CL_CheckBox* _stationary_check_box;

    CL_ScrollBar* _scale_scroll_bar;

    CL_InputBox* _mass_input_box;

    CL_CheckBox* _rotX_check_box;
    CL_CheckBox* _rotY_check_box;
    CL_CheckBox* _rotZ_check_box;

    map<string, CreatableObject > _creatable_objects;

    Object* _edited_object;
    //@}
};

#endif
