/**
 * @file input.cpp
 * Implementation of input handler.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */

#include "input.h"
#include "gui.h"
#include "gamestate.h"
#include "level.h"
#include "camera.h"
#include "object.h"

InputHandler::InputHandler( GUI* gui, GameState* gamestate ) :
    _gui( gui ),
    _gamestate( gamestate ),
    _ignore_move_signal( false )
{
    _control_vertical = false;
    _control_rotating = false;
}

InputHandler::~InputHandler( void )
{

}

bool InputHandler::_out_of_bounds( Vec3 newpos )
{
    if( !_gamestate || !_gamestate->get_level() ) {
        cerr << "InputHandler::_out_of_bounds: NULL gamestate/level!" << endl;
        return false;
    }

    Vec3 world_size = _gamestate->get_level()->get_world_size() / 2.0;

    return ( fabs( newpos[ 0 ] ) > world_size[ 0 ] ||
             fabs( newpos[ 1 ] ) > world_size[ 1 ] ||
             fabs( newpos[ 2 ] ) > world_size[ 2 ] );
}

void InputHandler::_put_in_bounds( Vec3& newpos )
{
    if( !_gamestate || !_gamestate->get_level() ) {
        cerr << "InputHandler::_put_in_bounds: NULL gamestate/level!" << endl;
        return;
    }

    Vec3 world_size = _gamestate->get_level()->get_world_size() / 2.0;

    for ( int i = 0; i < 3; i++ ) {
        if ( newpos[ i ] > world_size[ i ] ) {
            newpos[ i ] = world_size[ i ];
        }
        else if ( newpos[ i ] < -world_size[ i ] ) {
            newpos[ i ] = -world_size[ i ];
        }
    }
}

int InputHandler::register_signals( void )
{
    if ( !_gui || !_gamestate )
        return -1;

    _mouse_up = CL_Mouse::sig_key_up().connect( this, &InputHandler::_on_mouse_up );
    _mouse_down = CL_Mouse::sig_key_down().connect( this, &InputHandler::_on_mouse_down );
    _mouse_move = CL_Mouse::sig_move().connect( this, &InputHandler::_on_mouse_move );

    _key_up = CL_Keyboard::sig_key_up().connect( this, &InputHandler::_on_key_up );
    _key_down = CL_Keyboard::sig_key_down().connect( this, &InputHandler::_on_key_down );

    return 0;
}

void InputHandler::_on_mouse_up( const CL_InputEvent& event )
{
    if ( event.id == CL_MOUSE_RIGHT ) {
        _control_vertical = false;
    }
}

void InputHandler::_on_mouse_down( const CL_InputEvent& event )
{
    if ( !_gamestate || !_gui || _gamestate->get_current_state() != GameState::GAMEPLAY ) {
        cerr << "InputHandler::_on_mouse_down: invalid context!" << endl;
        return;
    }

    vector<ButtonHandler>& button_handlers = _gui->get_button_handlers();

    // check if a gui button was clicked on first, if not then do the
    // object-based input (so that buttons can't be clicked at the same time
    // objects are selected, which causes odd problems)
    for ( uint i = 0; i < button_handlers.size(); i++ ) {
        string name = button_handlers[ i ].name;
        CL_Component* button = _gui->get_button( name );

        if ( ( name == "StartSim" || name == "ResetSim" ) &&
             button->get_screen_rect().is_inside( event.mouse_pos ) ) {
            // clicked on a GUI button, doesn't count.
            return;
        }
    }

    if( !_gamestate ) {
        return;
    }

    // for picking
    _gamestate->get_glstate()->set_active();

    if ( event.id == CL_MOUSE_RIGHT ) {
        _control_vertical = true;
    }
    else {
        _control_vertical = false;
    }

    if ( _gamestate->get_level()->get_selected_object() ) {
        if ( event.id == CL_MOUSE_LEFT ) {
            // attempt to deselect the selected object
            int deselect_result = _gamestate->get_level()->deselect_object();

            if ( deselect_result == UPDATE_MOUSE ) {
                _update_mouse_pos();
            }
            else if ( deselect_result == COLLIDING ) {
                _play_sound();
            }
        }
    }
    else {
        int pos[] = { event.mouse_pos.x, event.mouse_pos.y };

        // select the object clicked on and move the mouse to its center
        _gamestate->get_level()->select_object( pos );
        _update_mouse_pos();
    }

    _prev_mouse_pos[ 0 ] = event.mouse_pos.x;
    _prev_mouse_pos[ 1 ] = event.mouse_pos.y;
}

void InputHandler::_play_sound( void )
{
    CL_SoundBuffer sample( "resources/sounds/error.wav" );
    sample.play();
}

void InputHandler::_on_mouse_move( const CL_InputEvent& event )
{
    // mouse moved by update_mouse_pos
    if ( _ignore_move_signal ) {
        _prev_mouse_pos[ 0 ] = event.mouse_pos.x;
        _prev_mouse_pos[ 1 ] = event.mouse_pos.y;
        _ignore_move_signal = false;
        return;
    }

    if( !_gamestate || !_gamestate->get_level() ) {
        cerr << "InputHandler::_on_mouse_move: NULL gamestate/level!" << endl;
        return;
    }

    if ( _gamestate->get_current_state() == GameState::GAMEPLAY ||
         _gamestate->get_current_state() == GameState::RUN_SIM ) {

        Vec2 cur_mouse_pos( event.mouse_pos.x, event.mouse_pos.y );
        Object *object = _gamestate->get_level()->get_selected_object();

        // translate/rotate the selected object
        if ( object && _gamestate->get_current_state() == GameState::GAMEPLAY ) {
            _gamestate->get_glstate()->set_active();

            _mouse_drag_object( _gamestate->get_level()->get_camera(), object, _prev_mouse_pos, cur_mouse_pos );

            _gamestate->get_level()->check_selected_colliding();
        }

        _prev_mouse_pos[ 0 ] = event.mouse_pos.x;
        _prev_mouse_pos[ 1 ] = event.mouse_pos.y;
    }
}

void InputHandler::_on_key_up( const CL_InputEvent &event )
{
    if( !_gamestate || !_gamestate->get_level() ) {
        cerr << "InputHandler::_on_key_up: NULL gamestate/level!" << endl;
        return;
    }

    if( _gamestate->get_current_state() == GameState::GAMEPLAY ||
        _gamestate->get_current_state() == GameState::RUN_SIM ) {

        // if the user is done rotating the object, make sure the mouse
        // is back at the object's center
        if ( event.id == CL_KEY_CONTROL ) {
            if ( _gamestate->get_level()->get_selected_object() ) {
                _update_mouse_pos();
            }
            _control_rotating = false;
        }
    }
}

void InputHandler::_on_key_down( const CL_InputEvent &event )
{
    if ( !_gui || !_gamestate || !_gamestate->get_level() ) {
        cerr << "InputHandler::_on_key_down: NULL gui/gamestate/level!" << endl;
        return;
    }

    switch( event.id ) {
    case CL_KEY_ESCAPE:
        // toggle the esc dialog if there is no object selected
        if ( ( _gamestate->get_current_state() == GameState::GAMEPLAY ||
               _gamestate->get_current_state() == GameState::ESC_DIALOG ) &&
              !_gamestate->get_level()->get_selected_object() ) {

            _gui->toggle_esc_dialog();
        }

        break;
    }

    // the rest of the key presses are only applicable in these two states
    if ( _gamestate->get_current_state() == GameState::GAMEPLAY ||
         _gamestate->get_current_state() == GameState::RUN_SIM ) {

        _key_down_camera( _gamestate->get_level()->get_camera() );

        Object *object = _gamestate->get_level()->get_selected_object();

        // fine-tune the object's position/orientation
        if ( object && _gamestate->get_current_state() == GameState::GAMEPLAY ) {
            _gamestate->get_glstate()->set_active();

            _key_down_object( _gamestate->get_level()->get_camera(), object );

            _gamestate->get_level()->check_selected_colliding();
        }

        switch ( event.id ) {
        case CL_KEY_ESCAPE:
            {
                int deselect_result = _gamestate->get_level()->deselect_object();

                if ( deselect_result == UPDATE_MOUSE ) {
                    _update_mouse_pos();
                }
                else if ( deselect_result == COLLIDING ) {
                    _play_sound();
                }

            }
            break;

        case CL_KEY_CONTROL:
            _control_rotating = true;
            break;

        case CL_KEY_DELETE:
            {
                Object* deleted_object = _gamestate->get_level()->get_selected_object();

                // level-editor command: remove the currently selected object
                // from the level
                if ( _gamestate->get_level()->is_editor_mode() && deleted_object ) {
                    _gamestate->get_level()->set_selected_object( NULL );

                    _gamestate->get_level()->remove_object( deleted_object );
                }
            }
            break;

        case CL_KEY_1:
            _gamestate->get_level()->toggle_show_movables();
            break;

        case CL_KEY_2:
            _gamestate->get_level()->set_axes_type( Level::AxesType(
                        (int(_gamestate->get_level()->get_axes_type()) + 1) %
                        int(Level::DRAW_NUM_TYPES) ) );
            break;

        case CL_KEY_E:
            _gamestate->get_level()->toggle_editor_mode();
            break;

        case CL_KEY_R:
            {
                Object* selected_object = _gamestate->get_level()->get_selected_object();

                // reset the selected object's orientation to the identity rotation
                if ( selected_object && ( _gamestate->get_level()->is_editor_mode() ||
                            ( selected_object->is_rotatable( 0 ) &&
                              selected_object->is_rotatable( 1 ) &&
                              selected_object->is_rotatable( 2 ) ) ) ) {

                    dQuaternion identity;

                    dQSetIdentity( identity );

                    selected_object->set_orientation_quat( identity );
                }
            }
            break;

        case CL_KEY_P:
            // level-editor command: bring up the "create object" menu
            if ( _gamestate->get_level()->is_editor_mode() ) {
                _gui->show_context_window();
            }
            break;

        case CL_KEY_T:
            // level-editor command: bring up the "edit object" menu
            if ( _gamestate->get_level()->is_editor_mode() &&
                 _gamestate->get_level()->get_selected_object() ) {

                _gui->edit_object( _gamestate->get_level()->get_selected_object() );
            }
            break;

        default:
            // update_mouse_pos();
            break;
        }
    }
}

void InputHandler::_update_mouse_pos( void )
{
    if ( !_gamestate || !_gamestate->get_level() ||
         !_gamestate->get_level()->get_selected_object() ) {
        // no object? don't care where the mouse is
        return;
    }

    _gamestate->get_glstate()->set_active();

    Vec3 pos = _gamestate->get_level()->get_selected_object()->get_position();
    Vec3 mouse = world_to_screen( pos );

    // move the mouse, but make sure we don't treat it as a move event.
    // NOTE: clanlib crashes if you disconnect the signal inside the
    // callback.
    _ignore_move_signal = true;

    CL_Mouse::set_position( int( mouse[0] ), int( mouse[1] ) );
}

void InputHandler::_mouse_drag_object( Camera *camera, Object *obj, Vec2 prev, Vec2 cur )
{
    if( !_gamestate || !_gamestate->get_level() ||
        !camera || !obj )
        return;

    bool editor_mode = _gamestate->get_level()->is_editor_mode();

    if ( _control_rotating ) {
        // rotation
        Quat quat;
        dQuaternion_to_quat( quat, dBodyGetQuaternion( obj->get_body_id() ) );

        double rot_angle = 2.0 * acos( quat.scalar() );
        Vec3 rot_axis = quat.vector();

        Vec3 up = camera->get_up_vec();
        Vec3 right = camera->get_right_vec();
        Vec3 forward = up ^ right;

        // invert the object's transform, so we rotate about the camera's
        // up vector.
        Mat4 inv_rot = rotation_matrix_rad( -rot_angle, rot_axis );

        Vec2 drag = 5.0 * Vec2( ( cur[ 0 ] - prev[ 0 ] ) / CL_Display::get_width(),
                                ( cur[ 1 ] - prev[ 1 ] ) / CL_Display::get_height() );

        if ( _control_vertical ) {
            Quat qy = axis_to_quat( inv_rot * up, drag[ 0 ] );

            for ( int i = 0; i < 3; i++ ) {
                if ( !obj->is_rotatable( i ) && !editor_mode ) {
                    qy.vector()[ i ] = 0;
                }
            }

            quat = quat * qy;
        }
        else {
            Quat qx = axis_to_quat( inv_rot * right, drag[ 1 ] );
            Quat qz = axis_to_quat( inv_rot * forward, drag[ 0 ] );

            for ( int i = 0; i < 3; i++ ) {
                if ( !obj->is_rotatable( i ) && !editor_mode ) {
                    qx.vector()[ i ] = 0;
                    qz.vector()[ i ] = 0;
                }
            }

            quat = quat * qz * qx;
        }
        unitize( quat );

        dQuaternion dq;
        quat_to_dQuaternion( dq, quat );

        obj->set_orientation_quat( dq );
    }
    else {
        // translation
        if( !obj->is_movable() && !editor_mode )
            return;

        // goal: try to keep object as close to the mouse as possible.
        // this is achieved by retrieving the mouse's position in the
        // world, then moving the object to that position (after
        // constraining movement to be in the appropriate axes).  due to
        // movement constraint, there will be some error, so we do it a
        // bunch of times the object hasn't moved much.
        for ( int i = 0; i < 10; i++ ) {
            Vec3 pos = obj->get_position();
            Vec3 mouse = world_to_screen( pos );
            mouse[ 0 ] = cur[ 0 ];
            mouse[ 1 ] = cur[ 1 ];

            Vec3 newpos = screen_to_world( mouse );

            if ( _control_vertical ) {
                // drag along y axis (ie, don't change x or z)
                newpos[ 0 ] = pos[ 0 ];
                newpos[ 2 ] = pos[ 2 ];
            }
            else {
                // drag in xz plane (ie, don't change y)
                newpos[ 1 ] = pos[ 1 ];
            }

            if ( _out_of_bounds( newpos ) ) {
                _put_in_bounds( newpos );
            }

            obj->set_position( newpos );

            if ( norm2( newpos - pos ) < 0.1 ) {
                // object didn't move much: we're close enough.
                break;
            }
        }
    }
}

void InputHandler::_key_down_object( Camera *camera, Object* obj )
{
    if( !_gamestate || !_gamestate->get_level() ||
        !camera || !obj )
        return;

    const double amount = 0.1;

    bool editor_mode = _gamestate->get_level()->is_editor_mode();
    Vec3 pos = obj->get_position();

    if ( obj->is_movable() || editor_mode ) {
        if ( CL_Keyboard::get_keycode( CL_KEY_S ) ) {
            obj->set_position( obj->get_position() + Vec3( 0.0, amount, 0.0 ) );
        }
        if ( CL_Keyboard::get_keycode( CL_KEY_X ) ) {
            obj->set_position( obj->get_position() + Vec3( 0.0, -amount, 0.0 ) );
        }
        if ( CL_Keyboard::get_keycode( CL_KEY_Z ) ) {
            obj->set_position( obj->get_position() + Vec3( -amount, 0.0, 0.0 ) );
        }
        if ( CL_Keyboard::get_keycode( CL_KEY_C ) ) {
            obj->set_position( obj->get_position() + Vec3( amount, 0.0, 0.0 ) );
        }
        if ( CL_Keyboard::get_keycode( CL_KEY_F ) ) {
            obj->set_position( obj->get_position() + Vec3( 0.0, 0.0, -amount ) );
        }
        if ( CL_Keyboard::get_keycode( CL_KEY_V ) ) {
            obj->set_position( obj->get_position() + Vec3( 0.0, 0.0, amount ) );
        }

        Vec3 pos = obj->get_position();
        if ( _out_of_bounds( pos ) ) {
            _put_in_bounds( pos );
            obj->set_position( pos );
        }
    }

    // rotation
    Quat quat;
    dQuaternion_to_quat( quat, dBodyGetQuaternion( obj->get_body_id() ) );

    if ( obj->is_rotatable( 0 ) || editor_mode ) {
        double x = 0.0;
        if ( CL_Keyboard::get_keycode( CL_KEY_K ) ) {
            x = amount;
        }
        if ( CL_Keyboard::get_keycode( CL_KEY_M ) ) {
            x = -amount;
        }
        if ( x != 0.0 ) {
            quat = quat * axis_to_quat( Vec3( 1.0, 0.0, 0.0 ), x );
        }
    }

    if ( obj->is_rotatable( 1 ) || editor_mode ) {
        double x = 0.0;
        if ( CL_Keyboard::get_keycode( CL_KEY_L ) ) {
            x = amount;
        }
        if ( CL_Keyboard::get_keycode( '.' ) ) {
            x = -amount;
        }
        if ( x != 0.0 ) {
            quat = quat * axis_to_quat( Vec3( 0.0, 1.0, 0.0 ), x );
        }
    }

    if ( obj->is_rotatable( 2 ) || editor_mode ) {
        double x = 0.0;
        if ( CL_Keyboard::get_keycode( ',' ) ) {
            x = amount;
        }
        if ( CL_Keyboard::get_keycode( '/' ) ) {
            x = -amount;
        }
        if ( x != 0.0 ) {
            quat = quat * axis_to_quat( Vec3( 0.0, 0.0, 1.0 ), x );
        }
    }

    dQuaternion dq;
    quat_to_dQuaternion( dq, quat );
    obj->set_orientation_quat( dq );
}

void InputHandler::_key_down_camera( Camera *camera )
{
    if( !camera )
        return;

    const double dangle = 5.0;
    const double dradius = 1.0;
    bool update = false;

    if ( CL_Keyboard::get_keycode( CL_KEY_LEFT ) ) {
        camera->set_azimuth( camera->get_azimuth() - dangle );
        update = true;
    }
    if ( CL_Keyboard::get_keycode( CL_KEY_RIGHT ) ) {
        camera->set_azimuth( camera->get_azimuth() + dangle );
        update = true;
    }
    if ( CL_Keyboard::get_keycode( CL_KEY_DOWN ) ) {
        camera->set_elevation( camera->get_elevation() - dangle );
        update = true;
    }
    if ( CL_Keyboard::get_keycode( CL_KEY_UP ) ) {
        camera->set_elevation( camera->get_elevation() + dangle );
        update = true;
    }
    if ( CL_Keyboard::get_keycode( CL_KEY_PRIOR ) ) {
        camera->set_radius( camera->get_radius() - dradius );
        update = true;
    }
    if ( CL_Keyboard::get_keycode( CL_KEY_NEXT ) ) {
        camera->set_radius( camera->get_radius() + dradius );
        update = true;
    }

    if( update ) {
        _update_mouse_pos();
    }
}
