/**
 * @file input.h
 * Implementation of the InputHandler class.
 *
 * Copyright 2003 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _INPUTHANDLER_H
#define _INPUTHANDLER_H

#include "main.h"
#include "gfxutils.h"

#include <ClanLib/core.h>
#include <ClanLib/gui.h>
#include <ClanLib/display.h>
#include <ClanLib/sound.h>

#include <gfx/vec2.h>
#include <gfx/mat4.h>

#include <ode/ode.h>

class GUI;
class GameState;
class Object;
class Camera;

/**
 * Handles all input from the keyboard and mouse.
 *
 * Passes messages to GUI and GameState resulting from input logic.
 *
 * @see GUI
 * @see GameState
 */
class InputHandler
{
public:
    /**
     * InputHandler constructor.
     *
     * @param gui       the game's gui
     * @param gamestate the game's gamestate
     */
    InputHandler( GUI* gui = NULL, GameState* gamestate = NULL );

    ~InputHandler( void );

    /**
     * Register the inputhandler's signal handlers with clanlib.
     *
     * @return 0 on success, -1 on failure
     */
    int register_signals( void );

private:

    //@{
    /** @name Mouse input
     *
     * Mouse input handlers.
     */
    void _on_mouse_up( const CL_InputEvent& event );
    void _on_mouse_down( const CL_InputEvent& event );
    void _on_mouse_move( const CL_InputEvent& event );
    //@}

    //@{
    /** @name Keyboard input
     *
     * Keyboard input handlers.
     */
    void _on_key_up( const CL_InputEvent& event );
    void _on_key_down( const CL_InputEvent& event );
    //@}

    /**
     * Update the mouse position based on where the selected object is.
     */
    void _update_mouse_pos( void );

    /**
     * Move the camera based on user input.
     *
     * @param camera the camera to update
     */
    void _key_down_camera( Camera *camera );

    /**
     * Move or rotate an object based on keyboard input.
     *
     * @param camera unused
     * @param obj    the object to move
     */
    void _key_down_object( Camera *camera, Object* obj );

    /**
     * Move or rotate an object based on mouse input.
     *
     * For a description of available commands see the help menu.
     *
     * @param camera use this camera's up/forward/right vectors for rotation
     * @param obj    the object to move
     * @param prev   the mouse's previous position
     * @param cur    the mouse's current position
     */
    void _mouse_drag_object( Camera *camera, Object *obj, Vec2 prev, Vec2 cur );

    /**
     * SHOULD NOT BE IN HERE, NEED A SOUND CLASS!!!
     */
    void _play_sound( void );

    /**
     * Check if a position is out of the bounds of the world.
     *
     * @param newpos the position to check against the world bounds
     *
     * @return true if out of the world, false otherwise
     *
     * @see InputHandler::mouse_drag_object
     */
    bool _out_of_bounds( Vec3 newpos );

    /**
     * Clamp a position to the world boundaries.
     *
     * @param newpos the position to clamp
     */
    void _put_in_bounds( Vec3& newpos );


    GUI* _gui;              ///< reference to the game's gui

    GameState* _gamestate;  ///< reference to the game's gamestate

    //@{
    /** @name Signal slots
     */
    CL_Slot _key_up;
    CL_Slot _key_down;

    CL_Slot _mouse_up;
    CL_Slot _mouse_down;
    CL_Slot _mouse_move;
    //@}


    bool _ignore_move_signal;   ///< so we can move the mouse manually without entering an infinite loop

    int _prev_mouse_pos[ 2 ];   ///< keep the mouse's previous position

    bool _control_vertical;  ///< whether we are moving the object vertical.
    bool _control_rotating;  ///< true if rotating the obj, false if moving.
};

#endif // header guard
