/**
 * @file main.cpp
 * Implementation of main application.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */


#include "main.h"
#include "gamestate.h"
#include "gui.h"
#include "input.h"
#include "model.h"
#include "material.h"

#include <iostream>
using namespace std;

#include <ClanLib/core.h>
#include <ClanLib/gui.h>
#include <ClanLib/gl.h>
#include <ClanLib/guistylesilver.h>
#include <ClanLib/application.h>
#include <ClanLib/display.h>
#include <ClanLib/sound.h>


#define CL_MODULES( func )   \
    CL_SetupCore::func();    \
    CL_SetupSound::func();   \
    CL_SetupDisplay::func(); \
    CL_SetupGL::func();      \
    CL_SetupGUI::func()

class AP : public CL_ClanApplication {
public:
    virtual int main( int argc, char** argv )
    {
        // Create a console window for text-output if not available
        CL_ConsoleWindow console("Console", 80, 10000 );
        console.redirect_stdio();

        CL_CrashReporter* crash_reporter = new CL_CrashReporter();
        crash_reporter->set_dump_to_file( true );

        try {
            cout << "Initializing ClanLib..." << endl;

            CL_MODULES( init );

            CL_SoundOutput output( 44100 );
            output.set_global_volume( 100 );

            cout << "Creating GUI..." << endl;

            _gui = new GUI();

            // FIXME: softcode this?
            Material::set_dir( "resources/textures" );
            Model::set_dir( "resources/models" );

            cout << "Adding Resources..." << endl;

            if( _gui->add_resources( "resources/gui/GUIStyleSilver/gui.xml" ) < 0 )
                return -1;

            if( _gui->add_resources( "resources/gui/gui.xml" ) < 0 )
                return -1;

            cout << "Initializing GUI..." << endl;

            if( _gui->initialize() < 0 )
                return -1;

            cout << "Registering GUI signals..." << endl;

            if( _gui->register_signals() < 0 )
                return -1;

            cout << "Creating GameState..." << endl;

            _gamestate = new GameState( _gui->get_gc() );

            _gui->set_gamestate( _gamestate );

            cout << "Creating Input Handler..." << endl;

            _input_handler = new InputHandler( _gui, _gamestate );

            cout << "Registering Input Handler signals..." << endl;

            if( _input_handler->register_signals() < 0 )
                return -1;

            cout << "Entering Game Loop..." << endl;

            _last_tick = CL_System::get_time();

            while( !_gui->quit() )
            {
                _cur_tick = CL_System::get_time();
                double t = double(_cur_tick - _last_tick) / 1000.0;

                _gamestate->tick( 0.01 );

                _last_tick = _cur_tick;

                _gamestate->render();

                _gui->render();

                CL_Display::flip();
                CL_System::keep_alive( 10 );
            }

            _gui->finish();

            dCloseODE();

            delete _gui;
            delete _gamestate;
            delete _input_handler;

            CL_MODULES( deinit );
        }
        catch ( CL_Error e ) {
            cout << e.message << endl;

            // Display console close message and wait for a key
            console.display_close_message();
        }

        return 0;
    }

private:

    GameState* _gamestate;

    GUI* _gui;

    InputHandler* _input_handler;

    int _last_tick, _cur_tick;

} app;
