/**
 * @file camera.h
 * Header for Camera class.
 *
 * Copyright 2003 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _CAMERA_H
#define _CAMERA_H

#include "main.h"
#include "gfxutils.h"

#include <math.h>
#include <iostream>
using namespace std;

#include <ClanLib/display.h>

#include <gfx/gl.h>

/**
 * A camera that always looks at the origin.
 *
 * Doesn't use true spherical coords, as phi is usually the angle from the
 * vertical axis, whereas the equivalent "elevation" is the angle from the
 * horizontal plane (basically, 90 - phi).  Also, all the axes are different
 * (e.g., y is up here, not z).
 *
 * @see Level
 */
class Camera {
public:

    /**
     * Different types of projections the camera supports.
     */
    enum ProjectType {
        ORTHOGRAPHIC,       ///< orthographic projection
        PERSPECTIVE         ///< perspective projection
    };

    /**
     * Camera constructor
     *
     * @param radius distance from the origin
     * @param elevation vertical angle the camera makes with the floor
     * @param azimuth horizontal angle "around" the y axis
     * @param pt type of projection to be used
     */
    Camera( double radius = 10.0, double elevation = 0.0, double azimuth = 0.0, ProjectType pt = PERSPECTIVE );

    //@{
    /** @name Accessors
     *
     * Accessor functions.
     */
    void set_radius( const double radius );
    void set_elevation( const double elevation );
    void set_azimuth( const double azimuth );

    double get_radius( void ) const { return _radius; }
    double get_elevation( void ) const { return _elevation; }
    double get_azimuth( void ) const { return _azimuth; }
    //@}

    //@{
    /** @name Derived Accessors
     *
     * Accessors to the camera's derived quantities.
     */
    Vec3r get_position( void ) const { return _position; }

    Vec3r get_right_vec( void ) const { return _right_vec; }
    Vec3r get_up_vec( void ) const { return _up_vec; }
    Vec3r get_forward_vec( void ) const { return _forward_vec; }
    //@}

    /**
     * Apply the camera's viewport and projection transformations.
     */
    void apply_projection( void );

    /**
     * Apply the camera's modelview transformations.
     */
    void apply_view( void );

    friend std::ostream& operator<<( std::ostream& os, const Camera& camera );

protected:

    /**
     * Update the camera's internal cache.
     */
    void _update( void );

    double _radius;              ///< distance of camera from origin
    double _elevation;           ///< vertical angle (phi in spherical coords)
    double _azimuth;             ///< horizontal angle (theta)

    ProjectType _project_type;   ///< type of projection

    //@{
    /** @name Limits
     *
     * Limits on the camera's position and orientation.
     */
    double _min_radius;
    double _max_radius;
    double _min_elevation;
    double _max_elevation;
    //@}

    //@{
    /** @name Derived quantities
     *
     * Derived quantities for use in other classes.
     *
     * @see Level::render
     * @see InputHandler::mouse_drag_object
     */
    Vec3r _position;
    Vec3r _right_vec;
    Vec3r _up_vec;
    Vec3r _forward_vec;
    //@}
};

#endif // header guard
