/**
 * @file scene/material.cpp
 * Implementation of Scene/material class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "material.h"
#include "utils.h"

#include <gfx/gl.h>
#include <ClanLib/core.h>
#include <ClanLib/display.h>

// wtf? why do i have to include this?
#include <ClanLib/Display/Providers/provider_type.h>

const Vec4f zero4f( 0.0f, 0.0f, 0.0f, 1.0f );

Material red(
        zero4f,
        Vec4f( 0.45f, 0.025f, 0.025f, 1.0f ),
        Vec4f( 0.45f, 0.025f, 0.025f, 1.0f ),
        Vec4f( 0.45f, 0.025f, 0.025f, 1.0f ),
        0.0f);

Material green(
        zero4f,
        Vec4f( 0.025f, 0.45f, 0.025f, 1.0f ),
        Vec4f( 0.025f, 0.45f, 0.025f, 1.0f ),
        Vec4f( 0.025f, 0.45f, 0.025f, 1.0f ),
        0.0f);

Material blue(
        zero4f,
        Vec4f( 0.025f, 0.025f, 0.45f, 1.0f ),
        Vec4f( 0.025f, 0.025f, 0.45f, 1.0f ),
        Vec4f( 0.025f, 0.025f, 0.45f, 1.0f ),
        0.0f);

Material grey(
        1.0f * Vec4f( 1.0f ),
        0.5f * Vec4f( 1.0f ),
        zero4f,
        0.5f * Vec4f( 1.0f ),
        0.0f);

string Material::_dir;

Material::Material(Vec4f a, Vec4f d, Vec4f s, Vec4f e, double shiny) :
    ambient(a), diffuse(d), specular(s), emissive(e), shininess(shiny),
    texture(NULL)
{
}

Material::Material( const Material& mat ) :
    texture( NULL )
{
    *this = mat;
}

const Material& Material::operator=( const Material& mat )
{
    if ( texture ) {
        delete texture;
    }

    ambient = mat.ambient;
    diffuse = mat.diffuse;
    specular = mat.specular;
    emissive = mat.emissive;
    shininess = mat.shininess;
    texture_filename = mat.texture_filename;
    texture = NULL;

    return *this;
}

Material::~Material( void )
{
    if ( texture ) {
        delete texture;
    }
}

Material Material::copy_with_texture( string filename ) const
{
    Material copy( *this );
    copy.texture_filename = filename;
    return copy;
}

void Material::reload_texture( void )
{
    if ( !texture_filename.empty() ) {
        texture_filename = convert_path( texture_filename );

        if ( texture ) {
            delete texture;
        }
        texture = load_gl_texture( get_dir() + PATH_SEP + texture_filename );
    }
}

void Material::make_current( void ) const
{
    const GLenum FACE = GL_FRONT_AND_BACK;

    // Copy this material's values into the current OpenGL material
    glMaterialfv( FACE, GL_AMBIENT, ambient );
    glMaterialfv( FACE, GL_DIFFUSE, diffuse );
    glMaterialfv( FACE, GL_SPECULAR, specular );
    glMaterialfv( FACE, GL_EMISSION, emissive );
    glMaterialf( FACE, GL_SHININESS, shininess );

    // for texture-warping (e.g. the conveyorbelt)
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );

    if ( texture ) {
        texture->bind();
    }
    else {
        // bind to 0 or the previous texture will be used for this material
        glBindTexture( GL_TEXTURE_2D, 0 );
    }
}

class CL_BMPProvider : public CL_ProviderType {
public:
    CL_BMPProvider() : CL_ProviderType("bmp") {}
    virtual CL_PixelBuffer *load( const string& filename, CL_InputSourceProvider* input_provider )
    {
        CL_InputSource *input = NULL;
        CL_PixelBuffer *buffer = NULL;
        try {
            if ( !input_provider ) {
                input_provider = CL_InputSourceProvider::create_file_provider( "." );
            }
            input = input_provider->open_source( filename );
            buffer = read_data( input );
        } catch ( CL_Error& error ) {
            if ( input ) delete input;
            if ( buffer ) delete buffer;
            throw error;
        }

        delete input;
        return buffer;
    };

    CL_PixelBuffer *read_data( CL_InputSource *input )
    {
        unsigned short int bfType;
        long int bfOffBits;
        short int biPlanes;
        short int biBitCount;
        long int biSizeImage;

        /* make sure the file is there and open it read-only (binary) */
        bfType = input->read_int16();
        /* check if file is a bitmap */
        if (bfType != 19778) {
            throw CL_Error("Not a Bitmap-File!");
        }        
        /* get the file size */
        /* skip file size and reserved fields of bitmap file header */
        input->seek( 8, CL_InputSource::seek_cur );

        /* get the position of the actual bitmap data */
        bfOffBits = input->read_int32();

        /* skip size of bitmap info header */
        input->seek( 4, CL_InputSource::seek_cur );

        /* get the width of the bitmap */
        int width = input->read_int32();

        /* get the height of the bitmap */
        int height = input->read_int32();

        /* get the number of planes (must be set to 1) */
        biPlanes = input->read_int16();
        if (biPlanes != 1) {
            throw CL_Error("number of Planes not 1!");
        }

        /* get the number of bits per pixel */
        biBitCount = input->read_int16();
        if (biBitCount != 24) {
            throw CL_Error("bits per pixel not 24");
            return 0;
        }

        /* calculate the size of the image in bytes */
        biSizeImage = width * height * 3;

        // might need to change format to bgr888.. but this works for me.
        CL_PixelBuffer *buffer = 
            new CL_PixelBuffer(width, height, width * 3, CL_PixelFormat::rgb888);
        buffer->lock();
        unsigned char *data = (unsigned char*)buffer->get_data();

        /* seek to the actual data */
        input->seek( bfOffBits, CL_InputSource::seek_set );

        input->read( data, biSizeImage );

        buffer->unlock();
        return buffer;
    }
};

CL_OpenGLSurface* load_gl_texture( string filename )
{
    static CL_BMPProvider* bmp_provider = NULL;
    if ( !bmp_provider ) {
        bmp_provider = new CL_BMPProvider;
    }

    CL_OpenGLSurface *image = NULL;
    try {
        image = new CL_OpenGLSurface( filename );

        return image;
    } catch ( CL_Error& error ) {
        if (image) delete image;

        cerr << "Error loading texture " << filename << ": " << error.message << endl;
        return NULL;
    }
}

