/**
 * @file material.h
 * Header file for Material class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _MATERIAL_H
#define _MATERIAL_H

#include "main.h"

#include <string>
using namespace std;

#include <ClanLib/gl.h>

#include <gfx/vec4.h>

/**
 * Material to be used in coloring/illuminating/texturing rendered objects.
 *
 * Consists of the usual phong constants plus a texture.
 */
class Material {
public:

    static void set_dir( string dir ) { _dir = dir; }
    static string get_dir() { return _dir; }

    /**
     * Default constructor.
     */
    Material( void ) : texture( NULL ) { }
    
    /**
     * Material constructor.
     *
     * @param amb   ambient coefficient
     * @param diff  diffuse coefficient
     * @param spec  specular coefficient
     * @param emit  emissive factor
     * @param shiny shininess factor
     */
    Material( Vec4f amb, Vec4f diff, Vec4f spec, Vec4f emit, double shiny );
    
    /**
     * Copy constructor.
     */
    Material( const Material& mat );

    /**
     * Overloaded equals operator.
     */
    const Material& operator=( const Material& mat );

    ~Material();

    /**
     * Overloaded copy constructor, adds a texture to the copied material.
     */
    Material copy_with_texture( string filename ) const;

    /**
     * Reload the material's texture.
     */
    void reload_texture( void );

    /**
     * Push the material into opengl to be used in rendering the next object
     */
    void make_current( void ) const;

    Vec4f ambient;      ///< fraction of ambient light reflected
    Vec4f diffuse;      ///< fraction of incoming light reflected diffusely
    Vec4f specular;     ///< fraction of incoming light reflected specularly
    Vec4f emissive;     ///< light emitted by this material

    double shininess;   ///< exponent for Phong illumination model

    CL_OpenGLSurface* texture;      ///< opengl texture index
    string texture_filename;        ///< filename of texture, for reloading

private:

    static string _dir;             ///< the directory textures reside in.
};

CL_OpenGLSurface* load_gl_texture( string filename );

// predefined materials
extern Material red;
extern Material green;
extern Material blue;
extern Material grey;

#endif // header guard
