/**
 * @file model.h
 * Header file for Model class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _MODEL_H
#define _MODEL_H

#include "main.h"
#include "utils.h"
#include "material.h"
#include "gfxutils.h"

#include <string>
using namespace std;

#define CUBE_FRONT  0
#define CUBE_BACK   1
#define CUBE_LEFT   2
#define CUBE_RIGHT  3
#define CUBE_TOP    4
#define CUBE_BOTTOM 5

/**
 * A graphical representation of an object.
 */
class Model {
public:
    static void set_dir( string dir ) { _dir = dir; }
    static string get_dir() { return _dir; }

    /**
     * Default constructor.
     */
    Model( void );

    /**
     * Model constructor.
     *
     * @param mat material of the model
     */
    Model( const Material& mat );

    virtual ~Model( void ) { }

    /**
     * Render the model to the screen.
     */
    virtual void render( void );

    /**
     * Draw selection decoration to let the user know the model is picked.
     *
     * @param color the RGB color of the selected decoration
     */
    virtual void draw_decor( Color color )
    { draw_wire_cube( get_min() * 1.02, get_max() * 1.02, color ); }

    //@{
    /** @name Accessors
     *
     * Accessor functions.
     */
    virtual Vec3r get_min( void ) { return Vec3r( -1.0 ); }
    virtual Vec3r get_max( void ) { return Vec3r( 1.0 ); }

    virtual string get_info()
        { return "texture=\"" + _material.texture_filename + "\""; }

    void set_material( Material m ) { _material = m; _material.reload_texture(); }
    //@}

    /**
     * Draw a wireframe axis-aligned box.
     *
     * @param min the bottom, left, rear corner of the box
     * @param max the top, right, front corner of the box
     * @param color the RGB color of the box
     */
    static void draw_wire_cube( const Vec3r& min, const Vec3r& max, Color color );

    /**
     * Draw a solid axis-aligned box.
     *
     * @param min the bottom, left, rear corner of the box
     * @param max the top, right, front corner of the box
     * @param draw_faces which faces of the cube to draw
     */
    static void draw_solid_cube( const Vec3r& min, const Vec3r& max, const bool draw_faces[6] = NULL );

    /**
     * Draw a capped cylinder aligned along the z axis.
     *
     * @param length the length of the cylinder
     * @param radius the radius of the spherical caps (and thus the cylinder as well)
     * @param quality how smoothly to render the cylinder
     */
    static void draw_capped_cylinder( double length, double radius, int quality = 3 );

    /**
     * Draw a cylinder.
     *
     * @param center center point of the cylinder
     * @param dir "long" axis of the cylinder
     * @param mat Material to use when rendering the cylinder
     * @param width diameter of the cylinder
     * @param height length of the cylinder
     */
    static void draw_cylinder( Vec3r center, Vec3r dir, const Material& mat, double width, double height );

private:

    static string _dir; ///< the directory model files reside in.

    Material _material; ///< the OpenGL material for this model.
};

#endif // header guard
