/**
 * @file msmodel.cpp
 * Implementation of MSModel class.
 *
 * Copyright 2003 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */

#include "msmodel.h"
#include "utils.h"

#include <ClanLib/gl.h>
#include <ClanLib/display.h>
#include <gfx/gl.h>
#include <fstream>

MSModel::MSModel( string filename )
{
    load_model( filename );
}

MSModel::~MSModel()
{
	if( _vertices ) {
		delete[] _vertices;
	}

	if( _triangles ) {
		delete[] _triangles;
	}

	if( _materials ) {
		delete[] _materials;
	}

	if( _meshes ) {
		delete[] _meshes;
	}

	if( _bones ) {
		delete[] _bones;
	}
}

void MSModel::render()
{
    for ( uint m = 0; m < _num_meshes; m++ ) {
        if ( _meshes[m].material_idx >= 0 ) {
            _materials[ _meshes[m].material_idx ].make_current();
        }
        else {
            // make sure there's no random textures hanging around
            grey.make_current();
            glBindTexture( GL_TEXTURE_2D, 0 );
        }

        glBegin( GL_TRIANGLES );
        for ( uint t = 0; t < _meshes[m].triangle_idx.size(); t++ ) {
            const Triangle& tri = _triangles[ _meshes[m].triangle_idx[t] ];
            for ( uint i = 0; i < 3; i++ ) {
                const Vertex& v = _vertices[ tri.vertex_idx[i] ];
                if ( v.bone_idx == -1 ) {
                    glTexCoord2f( tri.s[i], tri.t[i] );
                    glNormal3fv( tri.normals[i] );
                    glVertex3fv( v.pos );
                }
                else {
                    const Mat4& transform = _bones[ v.bone_idx ].current;

                    glTexCoord2f( tri.s[i], tri.t[i] );

                    Vec4 normal( transform * Vec4( tri.normals[i], 0.0 ) );
                    unitize( normal );
                    glNormal3dv( normal );

                    glVertex3dv( transform * Vec4( v.pos, 1.0 ) );
                }
            }
        }
        glEnd();
    }
}

void MSModel::tick( double t)
{
    _curtime += t;
    if ( _curtime > _time ) {
        _reset_animation();
    }

    for ( uint b = 0; b < _num_bones; b++ ) {
        Bone& bone = _bones[b];

        if ( bone.translate_frames.size() == 0 && bone.rotate_frames.size() == 0) {
            // no frames, no work to do
            continue;
        }

        Vec3r translate = _interp_keyframes( bone.cur_translate_frame, bone.translate_frames );
        Vec3r rotate = _interp_keyframes( bone.cur_rotate_frame, bone.rotate_frames );

        Mat4 transform = rotation_matrix_angles_rad( rotate );
        set_translation( transform, translate );

        // right multiply the relative transform by our keyframe transform.
        // this gives us the transformation in bone space.
        transform = bone.relative * transform;

        if ( bone.parent_idx == -1 ) {
            // for root, bone space is the same as absolute space
            bone.current = transform;
        }
        else {
            // for child, we have to apply our parents transformation after
            // this one, to get to absolute space.
            bone.current =
                _bones[ bone.parent_idx ].current * transform;
        }
    }
}

Vec3r MSModel::_interp_keyframes( int curframe, const vector<Keyframe>& keyframes )
{
    uint frame = curframe;
    while ( frame < keyframes.size()
            && keyframes[frame].time < _curtime ) {
        frame++;
    }

    if ( frame == 0 ) {
        // we just started, no interpolation.
        return keyframes[0].value;
    }
    else if ( frame == keyframes.size() ) {
        // we're ending, no interpolation.
        return keyframes[frame - 1].value;
    }
    else {
        // in between frame-1 and frame.
        const Keyframe& prevframe = keyframes[frame-1];
        const Keyframe& curframe = keyframes[frame];

        double diff = curframe.time - prevframe.time;
        double u = ( _curtime - prevframe.time ) / diff;

        return (float(1.0 - u)*prevframe.value + float(u)*curframe.value);
    }
}

void MSModel::reload_textures()
{
    for ( uint i = 0; i < _num_materials; i++ ) {
        _materials[i].reload_texture();
    }
}

void MSModel::_setup_bones()
{
    uint i;
    for ( i = 0; i < _num_bones; i++ ) {
        if ( _bones[i].parent_idx == -1 ) {
            // the root bone: A = R
            _bones[i].absolute = _bones[i].relative;
        }
        else {
            // a child: A = Aparent * R
            // note: because parent comes before child in _bones, we can be
            // sure that Aparent is already set up.
            _bones[i].absolute =
                _bones[ _bones[i].parent_idx ].absolute * _bones[i].relative;
        }
    }

    for ( i = 0; i < _num_vertices; i++ ) {
        // transform the vertices to their bone's coordinate system. this is
        // done so that rotating during animations rotates around the bone
        // center, not the world origin.
        if ( _vertices[i].bone_idx != -1 ) {
            Mat4 mat = inverse_t_r( _bones[ _vertices[i].bone_idx ].absolute );
            _vertices[i].pos = proj( mat * Vec4( _vertices[i].pos, 1.0 ) );
        }
    }

    for ( i = 0; i < _num_triangles; i++ ) {
        // transform normals as well.  use w = 0.0 so the normal doesn't
        // translate.
        for ( int j = 0; j < 3; j++ ) {
            const Vertex& v = _vertices[ _triangles[i].vertex_idx[j] ];
            if ( v.bone_idx != -1 ) {
                Mat4 mat = inverse_t_r( _bones[ v.bone_idx ].absolute );
                _triangles[i].normals[j] = proj( mat * Vec4( _triangles[i].normals[j], 0.0 ) );
            }
        }
    }

    _reset_animation();
}

void MSModel::_reset_animation()
{
    for ( uint b = 0; b < _num_bones; b++ ) {
        _bones[b].cur_translate_frame = 0;
        _bones[b].cur_rotate_frame = 0;
        _bones[b].current = _bones[b].absolute;
    }
    _curtime = 0.0;
}

void MSModel::_compare_bbox( Vec3r newpos )
{
    for ( int i = 0; i < 3; i++ ) {
        if ( newpos[i] < _min[i] ) {
            _min[i] = newpos[i];
        }
        if ( newpos[i] > _max[i] ) {
            _max[i] = newpos[i];
        }
    }
}

// Everything after here is for loading from the .ms3d file.

//@{
/**
 * These are Milkshape3D data structures.
 *
 * They are used to read in data from the MS3D file.
 */

// the structures must be packed, or the compiler will pad them with junk
// bytes.  that would screw things up when we try to load straight from the
// file into the structures.
#ifdef _MSC_VER
#pragma pack(push, packing)
#pragma pack(1)
#define PACKED
#else
#define PACKED   __attribute__((packed))
#endif

typedef char Byte;
typedef short Word;
typedef int Quad;

struct MSHeader
{
    char idstr[10];
    int version;
} PACKED;

struct MSVertex {
    Byte flags;
    float pos[3];
    char bone_idx;
    Byte refcount;
} PACKED;

struct MSTriangle {
    Word flags;
    Word vertex_idx[3];
    float normals[3][3];
    float s[3], t[3];
    Byte unused;    // i don't know what this does
    Byte mesh_idx;
} PACKED;

struct MSMaterial {
    char name[32];
    float ambient[4];
    float diffuse[4];
    float specular[4];
    float emissive[4];
    float shininess;    // 0 - 128
    float transparency; // 0 - 1
    Byte mode;
    char texture_filename[128];
    char alphamap[128];
} PACKED;

struct MSBone {
    Byte flags;
    char name[32];
    char parent_name[32];
    float rotate[3];
    float translate[3];
    Word num_rotate_frames;
    Word num_translate_frames;
} PACKED;

struct MSKeyframe {
    float time; // sec
    float value[3];
} PACKED;
//@}

#define READ(fin, type, data) \
    type data; \
    fin.read( (char*)&data, sizeof( data ) )

bool MSModel::load_model( string filename )
{
    _filename = filename;

    filename = get_dir() + PATH_SEP + _filename;

    ifstream fin( filename.c_str(), ios::in | ios::binary );
    if ( fin.fail() ) {
        cerr << "Error opening Milkshape3D model file: " << filename << endl;
        return false;
    }

    READ( fin, MSHeader, hdr );

    if ( strncmp(hdr.idstr, "MS3D", 4) != 0 ) {
        cerr << "Not a Milkshape3D model" << endl;
        return false;
    }

    if ( hdr.version < 3 || hdr.version > 4) {
        cerr << "Wrong Milkshape3D version: " << hdr.version << endl;
        return false;
    }

    int i;

    READ( fin, Word, num_vertices );
    _num_vertices = num_vertices;
    _vertices = new Vertex[ _num_vertices ];
    for ( i = 0; i < num_vertices; i++ ) {
        READ( fin, MSVertex, vertex );
        _vertices[i].bone_idx = vertex.bone_idx;
        _vertices[i].pos = Vec3r( vertex.pos );
        if ( i == 0 ) {
            _min = _max = Vec3r( vertex.pos );
        }
        else {
            _compare_bbox( Vec3r( vertex.pos ) );
        }

    }

    READ( fin, Word, num_triangles );
    _num_triangles = num_triangles;
    _triangles = new Triangle[ _num_triangles ];
    for ( i = 0; i < num_triangles; i++ ) {
        READ( fin, MSTriangle, triangle );
        _triangles[i].s = Vec3r(triangle.s);
        _triangles[i].t = Vec3r( 1.0f ) - Vec3r( triangle.t ); // ms3d is upside-down
        for ( int j = 0; j < 3; j++ ) {
            _triangles[i].normals[j] = Vec3r( triangle.normals[j] );
            _triangles[i].vertex_idx[j] = triangle.vertex_idx[j];

        }

    }

    READ( fin, Word, num_meshes );
    _num_meshes = num_meshes;
    _meshes = new Mesh[ _num_meshes ];
    for ( i = 0; i < num_meshes; i++ ) {
        fin.seekg( sizeof(Byte), ios::cur ); // flags
        fin.seekg( 32, ios::cur ); // name

        READ( fin, Word, num_triangles );
        _meshes[i].triangle_idx.resize( num_triangles );
        for ( int j = 0; j < num_triangles; j++ ) {
            READ( fin, Word, tri_idx );
            _meshes[i].triangle_idx[j] = (int)tri_idx;
        }

        READ( fin, Byte, mat_idx );

        _meshes[i].material_idx = mat_idx;
    }

    READ( fin, Word, num_materials );
    _num_materials = num_materials;
    _materials = new Material[ _num_materials ];
    for ( i = 0; i < num_materials; i++ ) {
        READ( fin, MSMaterial, material );

        _materials[i].ambient = Vec4f( material.ambient );
        _materials[i].diffuse = Vec4f( material.diffuse );
        _materials[i].specular = Vec4f( material.specular );
        _materials[i].emissive = Vec4f( material.emissive );
        _materials[i].shininess = material.shininess;
        _materials[i].texture_filename = material.texture_filename;
        _materials[i].texture = 0;
    }

    READ( fin, float, fps );

    fin.seekg( sizeof(float), ios::cur ); // current time

    READ( fin, Quad, num_frames );

    _time = double( num_frames ) / double( fps );

    READ( fin, Word, num_bones );
    _num_bones = num_bones;
    _bones = new Bone[ _num_bones ];

    map<string, int> bone_names;

    for ( i = 0; i < num_bones; i++ ) {
        READ( fin, MSBone, bone );

        if ( bone.parent_name[0] ) {
            // look up the parent bone index.
            // note: this depends on the fact that child bones always come
            // after their parents in the ms3d file.
            map<string, int>::iterator it;
            if ( ( it = bone_names.find( string( bone.parent_name ) ) )
                    == bone_names.end() ) {
                cerr << "Couldn't find parent bone: " << bone.parent_name << endl;
                return false;
            }
            _bones[i].parent_idx = (*it).second;
        }
        else {
            // must be the root bone.
            _bones[i].parent_idx = -1;
        }

        bone_names[bone.name] = i;

        // compose the tranformation of translate+rotate now.
        _bones[i].relative = rotation_matrix_angles_rad( Vec3r( bone.rotate ) );
        set_translation( _bones[i].relative, Vec3r( bone.translate ) );

        _bones[i].rotate_frames.resize( bone.num_rotate_frames );
        _bones[i].translate_frames.resize( bone.num_translate_frames );

        int j;
        for ( j = 0; j < bone.num_rotate_frames; j++ ) {
            READ( fin, MSKeyframe, msframe );
            Keyframe& frame = _bones[i].rotate_frames[j];
            frame.bone_idx = i;
            frame.time = msframe.time;
            frame.value = Vec3r(msframe.value);
        }

        for ( j = 0; j < bone.num_translate_frames; j++ ) {
            READ( fin, MSKeyframe, msframe );
            Keyframe& frame = _bones[i].translate_frames[j];
            frame.bone_idx = i;
            frame.time = msframe.time;
            frame.value = Vec3r(msframe.value);
        }
    }

    fin.close();

    reload_textures();
    _setup_bones();

    return true;
}
