/**
 * @file msmodel.h
 * Header for MSModel class.
 *
 * Copyright 2003 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _MSMODEL_H
#define _MSMODEL_H

#include "main.h"
#include "model.h"
#include "gfxutils.h"

#include <string>
#include <vector>
using namespace std;

#include <gfx/mat4.h>

/**
 * A Milkshape3D graphical model.
 */
class MSModel : public Model {
public:

    /**
     * A triangle mesh, composed of a bunch of triangles, and a material.
     */
    struct Mesh {
        int material_idx;
        vector<int> triangle_idx;
    };

    /**
     * A triangle and its vertices' normal values, as well as texture
     * coords.
     */
    struct Triangle {
        int vertex_idx[3];
        Vec3r normals[3];
        Vec3r s, t;
    };

    /**
     * A single vertex and the bone it's attached to.
     */
    struct Vertex {
        Vec3r pos;
        char bone_idx;
    };

    /**
     * Keyframe for animation.
     */
    struct Keyframe {
        int bone_idx;
        float time; ///< sec
        Vec3r value;
    };

    /**
     * A skeletal bone, needed for skeletal animation.
     */
    struct Bone {
        Mat4 relative;
        Mat4 absolute;
        Mat4 current;   ///< the one that is applied when drawing

        vector<Keyframe> translate_frames;
        vector<Keyframe> rotate_frames;

        int cur_translate_frame;
        int cur_rotate_frame;

        int parent_idx;
    };

    MSModel( string filename );

    virtual ~MSModel( void );

    virtual void render( void );

    /**
     * Advance the animation.
     * This involves interpolating between the two closest keyframes to our
     * current time, and updating our bones to the correct
     * location/orientation.
     *
     * @param t amount by which to advance the animation.
     */
    virtual void tick( double t );

    /**
     * Load in the textures from their files.
     */
    void reload_textures( void );

    /**
     * Load in the MS3D model from a model file.
     *
     * @param filename relative path + filename of the model file to be loaded
     */
    bool load_model( string pathname );

    //@{
    /** @name Accessors
     *
     * Accessor functions.
     */
    virtual Vec3r get_min( void ) { return _min; }
    virtual Vec3r get_max( void ) { return _max; }

    virtual string get_info() { return "model=\"" + _filename + "\""; }

    Vertex* get_vertices( void ) { return _vertices; }
    Triangle* get_triangles( void ) { return _triangles; }

    int get_num_vertices( void ) { return _num_vertices; }
    int get_num_triangles( void ) { return _num_triangles; }
    //@}

private:

    /**
     * Prepare the bone structures (only called once).
     *
     * This fills in the bone matrices, and transforms all vertices in
     * the model to their own bone's space so animation changes them around
     * their bone.  Transformation back to world coordinates happens in
     * render().
     */
    void _setup_bones( void );

    /**
     * Reset the animation.
     */
    void _reset_animation( void );

    /**
     * Interpolate between keyframes of the model's animation.
     *
     * @return the interpolation between the two keyframes we are between.
     */
    Vec3r _interp_keyframes( int curframe, const vector<Keyframe>& keyframes );

    /**
     * Adjust min,max values for our bbox based on the new vertex 'newpos'.
     *
     * @param newpos a new vertex to consider for our bounding box.
     */
    void _compare_bbox( Vec3r newpos );

protected:

    string _filename;               ///< the model's filename
    Material* _materials;           ///< list of materials used by the model
    uint _num_materials;
    Mesh* _meshes;                  ///< list of meshes that make up the model
    uint _num_meshes;
    Triangle* _triangles;           ///< list of triangles used in the mesh
    uint _num_triangles;
    Vertex* _vertices;              ///< list of vertices used in the mesh
    uint _num_vertices;
    Bone* _bones;                   ///< list of bones in the skeleton
    uint _num_bones;
    double _time;                   ///< total animation time, sec
    double _curtime;                ///< current time
    Vec3r _min;                      ///< minimum x,y,z (for bbox)
    Vec3r _max;                      ///< maximum x,y,z (for bbox)
};

#endif // header guard
