/**
 * @file utils.cpp
 * Implementation of Utils class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "main.h"
#include "utils.h"

#include <math.h>
#include <stdio.h>
#include <gfx/gl.h>

#include <ClanLib/core.h>
#include <ClanLib/display.h>
#include <ClanLib/gl.h>

Vec3r world_to_screen( Vec3r world )
{
    Vec3r screen;
    double mview[ 16 ], proj[ 16 ];
    int view[ 4 ];

    glGetDoublev( GL_MODELVIEW_MATRIX, mview ),
    glGetDoublev( GL_PROJECTION_MATRIX, proj ),
    glGetIntegerv( GL_VIEWPORT, view ),

    gluProject( world[ 0 ], world[ 1 ], world[ 2 ],
                mview, proj, view,
                &screen[ 0 ], &screen[ 1 ], &screen[ 2 ] );

    // y coord is upside-down
    screen[ 1 ] = CL_Display::get_height() - screen[ 1 ];

    return screen;
}

Vec3r screen_to_world( Vec3r screen )
{
    Vec3r world;
    double mview[ 16 ], proj[ 16 ];
    int view[ 4 ];

    glGetDoublev( GL_MODELVIEW_MATRIX, mview ),
    glGetDoublev( GL_PROJECTION_MATRIX, proj ),
    glGetIntegerv( GL_VIEWPORT, view ),

    // y coord is upside-down
    screen[ 1 ] = CL_Display::get_height() - screen[ 1 ];

    gluUnProject( screen[ 0 ], screen[ 1 ], screen[ 2 ],
                  mview, proj, view,
                  &world[ 0 ], &world[ 1 ], &world[ 2 ] );

    return world;
}

void set_translation( Mat4& mat, const Vec3r& translate )
{
    mat( 0, 3 ) = translate[ 0 ];
    mat( 1, 3 ) = translate[ 1 ];
    mat( 2, 3 ) = translate[ 2 ];
}

// Milkshape3D likes to store their rotations as (rx, ry, rz) triples,
// where each is an angle to rotate around that axis.
Mat4 rotation_matrix_angles_rad( const Vec3r& angles )
{
    double cr = cos( angles[ 0 ] );
    double sr = sin( angles[ 0 ] );
    double cp = cos( angles[ 1 ] );
    double sp = sin( angles[ 1 ] );
    double cy = cos( angles[ 2 ] );
    double sy = sin( angles[ 2 ] );
    double srsp = sr * sp;
    double crsp = cr * sp;

    Mat4 mat;

    mat( 0, 0 ) = cp * cy;
    mat( 1, 0 ) = cp * sy;
    mat( 2, 0 ) = -sp;

    mat( 0, 1 ) = srsp * cy - cr * sy;
    mat( 1, 1 ) = srsp * sy + cr * cy;
    mat( 2, 1 ) = sr * cp;

    mat( 0, 2 ) = crsp * cy + sr * sy;
    mat( 1, 2 ) = crsp * sy - sr * cy;
    mat( 2, 2 ) = cr * cp;

    mat( 3, 3 ) = 1.0;

    return mat;
}

Mat4 inverse_t_r( const Mat4& mat )
{
    // T = translation matrix, R = rotation matrix, tv = T.col(3);
    // (TR)^-1 = R^-1 * T^-1 = Rt * T^-1 = [Rt, -tv]
    Vec3r translate = proj( mat.col( 3 ) );

    Mat4 inv = transpose( mat );
    inv[ 3 ] = Vec4( 0.0, 0.0, 0.0, 1.0 );
    set_translation( inv, inv * -translate );

    return inv;
}

string convert_path( string path )
{
    string::size_type pos = 0;

    while ( ( pos = path.find( BAD_PATH_SEP, 0 ) ) != string::npos ) {
        path[ pos ] = PATH_SEP;
    }

    return path;
}

Shape get_shape( const CL_DomElement& element, string name )
{
    string shape_str = get_string( element, name );

    if ( shape_str == "box" ) {
        return SHAPE_BOX;
    }
    else if ( shape_str == "sphere" ) {
        return SHAPE_SPHERE;
    }
    else if ( shape_str == "cylinder" ) {
        return SHAPE_CYLINDER;
    }

    return SHAPE_BOX;
}
