/**
 * @file utils.h
 * Header file for Utils class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _UTILS_H
#define _UTILS_H

#include "main.h"
#include "shape.h"
#include "gfxutils.h"

#include <string>
using namespace std;

#include <ClanLib/core.h>

#include <gfx/mat4.h>

class Material;

#define DEG_TO_RAD(angle) (angle*M_PI/180.0)
#define RAD_TO_DEG(angle) (angle*180.0/M_PI)

/**
 * Convert world coordinates to screen coordinates
 *
 * @param world world coordinates to be converted
 */
Vec3r world_to_screen( Vec3r world );

/**
 * Convert screen coordinates to world coordinates
 *
 * @param screen screen coordinates to be converted
 */
Vec3r screen_to_world( Vec3r screen );

/**
 * Make a rotation matrix out of a triplet of angles to rotate around each
 * axis.
 * Milkshape3D likes to store their rotations this way.
 */
Mat4 rotation_matrix_angles_rad( const Vec3r& angles );

/**
 * Set the translation vectors for a transformation matrix.
 */
void set_translation( Mat4& mat, const Vec3r& translate );

/**
 * Invert a translation*rotation matrix.
 * Optimized inverse for T*R matrix.
 */
Mat4 inverse_t_r( const Mat4& mat );

//@{
/** @name CL_DomElement accessors
  *
  * Utility functions for grabbing a particular type of attribute from a
  * CL_DomElement
  *
  * @see Level::load
  */
inline string get_string( const CL_DomElement& element, string name )
{
    return element.get_attribute( name );
}

inline int get_int( const CL_DomElement& element, string name )
{
    return CL_String::to_int( element.get_attribute( name ) );
}

inline float get_float( const CL_DomElement& element, string name )
{
    return CL_String::to_float( element.get_attribute( name ) );
}

inline double get_double( const CL_DomElement& element, string name )
{
    return CL_String::to_double( element.get_attribute( name ) );
}

inline bool get_bool( const CL_DomElement& element, string name )
{
    return CL_String::to_bool( element.get_attribute( name ) );
}

inline Vec3r get_vec3( const CL_DomElement& element, string name )
{
    vector<string> tokens =
        CL_String::tokenize( element.get_attribute( name ), " " );

    return Vec3r( CL_String::to_double( tokens[ 0 ] ),
            CL_String::to_double( tokens[ 1 ] ),
            CL_String::to_double( tokens[ 2 ] ) );
}

inline CL_DomElement get_element( CL_ResourceManager& manager, string resource )
{
    return manager.get_resource( resource ).get_element();
}
//@}

/**
 * Component-wise multiplication of Vec's
 */
inline Vec3r vec_mult( Vec3r a, Vec3r b )
{
    return Vec3r( a[0] * b[0], a[1] * b[1], a[2] * b[2] );
}

/**
 * Convert a Vec4 to a Vec3r without a 'true' projection (just drop the 4th
 * component)
 */
inline Vec3r vec_dropw( Vec4 a )
{
    return Vec3r( a[ 0 ], a[ 1 ], a[ 2 ] );
}

//@{
/** @name CL_String conversion utilities
 *
 * Utility functions for converting from a particular type to a CL_String.
 */
inline string to_string( bool x )
{
    return CL_String::from_bool( x );
}

inline string to_string( int x )
{
    return CL_String::from_int( x );
}

inline string to_string( double x )
{
    return CL_String::from_double( x );
}

inline string to_string( Vec3r v )
{
    return CL_String::from_double(v[0]) + " " +
        CL_String::from_double(v[1]) + " " +
        CL_String::from_double(v[2]);
}
//@}

/**
 * Return the Shape attribute from an xml element.
 */
Shape get_shape( const CL_DomElement& element, string name );

/**
 * convert a path to the native system format
 *
 * @param path path to be converted
 *
 * @return path with appropriate forward/backward slashes
 */
string convert_path( string path );

#endif // header guard
